"""
Module handling the potential class.
"""
import numpy as np
from sarkas.potentials.force_pm import force_optimized_green_function as gf_opt
from sarkas.potentials import force_pm, force_pp
import fdint


class Potential:
    """
    Parameters specific to potential choice.

    Attributes
    ----------
    matrix : array
        Potential's parameters.

    method : str
        Algorithm to use for force calculations.
        "PP" = Linked Cell List (default).
        "P3M" = Particle-Particle Particle-Mesh.

    rc : float
        Cutoff radius.

    type : str
        Interaction potential: LJ, Yukawa, EGS, Coulomb, QSP, Moliere.

    pppm_aliases : array, shape(3)
        Number of aliases in each direction.

    pppm_cao : int
        Charge assignment order.

    on : bool
        Flag.

    pppm_mesh : array, shape(3), int
        Number of mesh point in each direction.

    Mx : int
        Number of mesh point along the :math:`x` axis.

    My : int
        Number of mesh point along the :math:`y` axis.

    Mz : int
        Number of mesh point along the :math:`z` axis.

    mx_max : int
        Number of aliases along the reciprocal :math:`x` direction.

    my_max : int
        Number of aliases along the reciprocal :math:`y` direction.

    mz_max : int
        Number of aliases along the reciprocal :math:`z` direction.

    G_ew : float
        Ewald parameter.

    G_k : array
        Optimized Green's function.

    hx : float
        Mesh spacing in :math:`x` direction.

    hy : float
        Mesh spacing in :math:`y` direction.

    hz : float
        Mesh spacing in :math:`z` direction.

    PP_err : float
        Force error due to short range cutoff.

    PM_err : float
        Force error due to long range cutoff.

    F_err : float
        Total force error.

    kx_v : array
        Array of :math:`k_x` values.

    ky_v : array
        Array of :math:`k_y` values.

    kz_v : array
        Array of :math:`k_z` values.

    """

    def __init__(self):
        self.type = "yukawa"
        self.method = "PP"
        self.matrix = None
        self.force_error = None
        self.measure = False
        self.box_lengths = 0.0
        self.box_volume = 0.0
        self.fourpie0 = 0.0
        self.QFactor = 0.0
        self.total_net_charge = 0.0
        self.pppm_on = False

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = 'Potential( \n'
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ')'
        return disp

    def from_dict(self, input_dict: dict) -> None:
        """
        Update attributes from input dictionary.

        Parameters
        ----------
        input_dict: dict
            Dictionary to be copied.

        """

        self.__dict__.update(input_dict)

    def setup(self, params) -> None:
        """Setup the potential class.

        Parameters
        ----------
        params: sarkas.core.Parameters
            Simulation's parameters.

        """
        # Check for cutoff radius
        if not self.type.lower() == 'fmm':
            self.linked_list_on = True  # linked list on
            if not hasattr(self, "rc"):
                print("\nWARNING: The cut-off radius is not defined. L/2 = {:1.4e} will be used as rc".format(
                    0.5 * params.box_lengths.min()))
                self.rc = params.box_lengths.min() / 2.
                self.linked_list_on = False  # linked list off

            if self.rc > params.box_lengths.min() / 2.:
                print("\nWARNING: The cut-off radius is > L/2. L/2 = ", params.box_lengths.min() / 2,
                      "will be used as rc")
                self.rc = params.box_lengths.min() / 2.
                self.linked_list_on = False  # linked list off

        # Check for electrons as dynamical species
        if self.type.lower() == 'qsp':
            mask = params.species_names == 'e'
            self.electron_temperature = params.species_temperatures[mask]
            params.ne = float(params.species_num_dens[mask])
            params.electron_temperature = float(params.species_temperatures[mask])
            params.qe = float(params.species_charges[mask])
            params.me = float(params.species_masses[mask])
        else:
            params.ne = params.species_charges.transpose() @ params.species_concentrations * params.total_num_density / params.qe

            # Check electron properties
            if hasattr(self, "electron_temperature_eV"):
                self.electron_temperature = params.eV2K * self.electron_temperature_eV
            if hasattr(self, "electron_temperature"):
                params.electron_temperature = self.electron_temperature
            else:
                # if the electron temperature is not defined. The total ion temperature will be used for it.
                # print("\nWARNING: electron temperature not defined. I will use the total ion temperature.")
                params.electron_temperature = params.total_ion_temperature
                self.electron_temperature = params.electron_temperature

        self.calc_electron_properties(params)

        if hasattr(self, "kappa"):
            if self.electron_temperature != params.total_ion_temperature :
                print("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! WARNING !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                print("You have defined both kappa and the electron_temperature. kappa's value will be used.")
                print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n")
            # Thomas-Fermi Length
            params.lambda_TF = params.a_ws / self.kappa

        # Update potential-specific parameters
        # Coulomb potential
        if self.type.lower() == "coulomb":
            from sarkas.potentials import coulomb
            coulomb.update_params(self, params)
        # Yukawa potential
        if self.type.lower() == "yukawa":
            from sarkas.potentials import yukawa
            yukawa.update_params(self, params)
        # exact gradient-corrected screening (EGS) potential
        if self.type.lower() == "egs":
            from sarkas.potentials import egs
            egs.update_params(self, params)
        # Lennard-Jones potential
        if self.type.lower() == "lj":
            from sarkas.potentials import lennardjones as lj
            lj.update_params(self, params)
        # Moliere potential
        if self.type.lower() == "moliere":
            from sarkas.potentials import moliere
            moliere.update_params(self, params)
        # QSP potential
        if self.type.lower() == "qsp":
            from sarkas.potentials import qsp
            qsp.update_params(self, params)

        # Compute pppm parameters
        if self.method == 'P3M' or self.method.lower() == 'pppm':
            self.pppm_on = True
            self.pppm_setup(params)

        # Copy needed parameters
        self.box_lengths = np.copy(params.box_lengths)
        self.box_volume = params.box_volume
        self.fourpie0 = params.fourpie0
        self.QFactor = params.QFactor
        self.total_net_charge = params.total_net_charge
        self.measure = params.measure

    @staticmethod
    def calc_electron_properties(params):
        """Calculate electronic parameters. See DFT notes on website.

        Parameters
        ----------
        params: sarkas.core.Parameters
            Simulation's parameters.

        """

        twopi = 2.0 * np.pi
        spin_degeneracy = 2.0  # g in the notes

        # FDINT calculates the I integrals not the F integrals see notes.
        fdint_fdk_vec = np.vectorize(fdint.fdk)
        fdint_ifd1h_vec = np.vectorize(fdint.ifd1h)

        # Inverse temperature for convenience
        beta_e = 1. / (params.kB * params.electron_temperature)

        # Plasma frequency
        params.electron_plasma_frequency = np.sqrt(
            4.0 * np.pi * params.qe ** 2 * params.ne / (params.fourpie0 * params.me))

        # de Broglie wavelength
        params.lambda_deB = np.sqrt(twopi * params.hbar2 * beta_e / params.me)
        lambda3 = params.lambda_deB ** 3

        # Landau length 4pi e^2 beta. The division by fourpie0 is needed for MKS units
        params.landau_length = 4.0 * np.pi * params.qe**2 * beta_e/params.fourpie0

        # chemical potential of electron gas/(kB T), obtained by inverting the density equation.
        params.eta_e = fdint_ifd1h_vec(lambda3 * np.sqrt(np.pi) * params.ne / 4.0)

        # Thomas-Fermi length obtained from compressibility. See eq.(10) in Ref. [3]_
        lambda_TF_sq = lambda3 / params.landau_length
        lambda_TF_sq /= (spin_degeneracy / np.sqrt(np.pi) * fdint_fdk_vec(k=-0.5, phi=params.eta_e))
        params.lambda_TF = np.sqrt(lambda_TF_sq)

        # Electron WS radius
        params.ae_ws = (3.0/(4.0 * np.pi * params.ne))**(1./3.)
        # Brueckner parameters
        params.rs = params.ae_ws / params.a0
        # Fermi wave number
        params.kF = (3.0 * np.pi ** 2 * params.ne) ** (1. / 3.)

        # Fermi energy
        params.fermi_energy = params.hbar2 * params.kF ** 2 / (2.0 * params.me)

        # Other electron parameters
        params.electron_degeneracy_parameter = params.kB * params.electron_temperature / params.fermi_energy
        params.relativistic_parameter = params.hbar * params.kF / (params.me * params.c0)

        # Eq. 1 in Murillo Phys Rev E 81 036403 (2010)
        params.electron_coupling = params.qe**2/(
                params.fourpie0 * params.fermi_energy * params.ae_ws* np.sqrt(params.electron_degeneracy_parameter**2))

        # Warm Dense Matter Parameter, Eq.3 in Murillo Phys Rev E 81 036403 (2010)
        params.wdm_parameter = 2.0/(params.electron_degeneracy_parameter + 1.0/params.electron_degeneracy_parameter)
        params.wdm_parameter *= 2.0/(params.electron_coupling + 1.0/params.electron_coupling)

        if params.magnetized:
            if params.units == 'cgs':
                params.electron_cyclotron_frequency = params.qe * np.linalg.norm(params.magnetic_field)/params.c0/params.me
            else:
                params.electron_cyclotron_frequency = params.qe * np.linalg.norm(params.magnetic_field) / params.me

            params.electron_magnetic_energy = params.hbar * params.electron_cyclotron_frequency
            tan_arg = 0.5 * params.hbar * params.electron_cyclotron_frequency * beta_e

            # Perpendicular correction
            params.horing_perp_correction = (params.electron_plasma_frequency/params.electron_cyclotron_frequency)**2
            params.horing_perp_correction *= (1.0 - tan_arg/np.tanh(tan_arg))
            params.horing_perp_correction += 1

            # Parallel correction
            params.horing_par_correction = 1 - (params.hbar * beta_e * params.electron_plasma_frequency)**2 / 12.

            # Quantum Anisotropy Parameter
            params.horing_delta = (params.horing_perp_correction - 1)
            params.horing_delta += (params.hbar* beta_e * params.electron_cyclotron_frequency)**2/12
            params.horing_delta /= params.horing_par_correction

    def update_linked_list(self, ptcls):
        """
        Calculate the pp part of the acceleration.

        Parameters
        ----------
        ptcls: sarkas.core.Particles
            Particles data.

        """
        ptcls.potential_energy, ptcls.acc = force_pp.update(ptcls.pos, ptcls.id, ptcls.masses, self.box_lengths,
                                           self.rc, self.matrix, self.force,
                                           self.measure, ptcls.rdf_hist)

        if not (self.type == "LJ"):
            # Mie Energy of charged systems
            # J-M.Caillol, J Chem Phys 101 6080(1994) https: // doi.org / 10.1063 / 1.468422
            dipole = ptcls.charges @ ptcls.pos
            ptcls.potential_energy += 2.0 * np.pi * np.sum(dipole ** 2) / (3.0 * self.box_volume * self.fourpie0)

    def update_brute(self, ptcls):
        """
        Calculate particles' acceleration and potential brutally.

        Parameters
        ----------
        ptcls: sarkas.core.Particles
            Particles data.

        """
        ptcls.potential_energy, ptcls.acc = force_pp.update_0D(ptcls.pos, ptcls.id, ptcls.masses, self.box_lengths,
                                               self.rc, self.matrix, self.force,
                                               self.measure, ptcls.rdf_hist)
        if not (self.type == "LJ"):
            # Mie Energy of charged systems
            # J-M.Caillol, J Chem Phys 101 6080(1994) https: // doi.org / 10.1063 / 1.468422
            dipole = ptcls.charges @ ptcls.pos
            ptcls.potential_energy += 2.0 * np.pi * np.sum(dipole ** 2) / (3.0 * self.box_volume * self.fourpie0)

    def update_pm(self, ptcls):
        """Calculate the pm part of the potential and acceleration.

        Parameters
        ----------
        ptcls : sarkas.core.Particles
            Particles' data

        """
        U_long, acc_l_r = force_pm.update(ptcls.pos, ptcls.charges, ptcls.masses,
                                          self.pppm_mesh, self.box_lengths, self.pppm_green_function,
                                          self.pppm_kx,
                                          self.pppm_ky,
                                          self.pppm_kz, self.pppm_cao)
        # Ewald Self-energy
        U_long += self.QFactor * self.pppm_alpha_ewald / np.sqrt(np.pi)
        # Neutrality condition
        U_long += - np.pi * self.total_net_charge ** 2.0 / (2.0 * self.box_volume * self.pppm_alpha_ewald ** 2)

        ptcls.potential_energy += U_long

        ptcls.acc += acc_l_r

    def update_pppm(self, ptcls):
        """Calculate particles' potential and accelerations using pppm method.

        Parameters
        ----------
        ptcls : sarkas.core.Particles
            Particles' data

        """
        self.update_linked_list(ptcls)
        self.update_pm(ptcls)

    def pppm_setup(self, params):
        """Calculate the P3M parameters.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters

        """
        # Change lists to numpy arrays for Numba compatibility
        if not isinstance(self.pppm_mesh, np.ndarray):
            self.pppm_mesh = np.array(self.pppm_mesh)

        if not isinstance(self.pppm_aliases, np.ndarray):
            self.pppm_aliases = np.array(self.pppm_aliases)

        # P3M parameters
        self.pppm_h_array = params.box_lengths / self.pppm_mesh
        
        self.matrix[-1, :, :] = self.pppm_alpha_ewald
        # Pack constants together for brevity in input list 
        kappa = 1. / params.lambda_TF if self.type == "Yukawa" else 0.0
        constants = np.array([kappa, self.pppm_alpha_ewald, params.fourpie0])
        # Calculate the Optimized Green's Function
        self.pppm_green_function, self.pppm_kx, self.pppm_ky, self.pppm_kz, params.pppm_pm_err = gf_opt(
            params.box_lengths, self.pppm_mesh, self.pppm_aliases, self.pppm_cao, constants)
        
        # Complete PM Force error calculation
        params.pppm_pm_err *= np.sqrt(params.total_num_ptcls) * params.a_ws ** 2 * params.fourpie0
        params.pppm_pm_err /= params.box_volume ** (2. / 3.)

        # Total Force Error
        params.force_error = np.sqrt(params.pppm_pm_err ** 2 + params.pppm_pp_err ** 2)

        self.force_error = params.force_error

    # def update_fmm(ptcls, params):
    #     """
    #
    #     Parameters
    #     ----------
    #     ptcls
    #     params
    #
    #     Returns
    #     -------
    #
    #     """
    #
    #     if params.potential.type == 'Coulomb':
    #         out_fmm = fmm.lfmm3d(eps=1.0e-07, sources=np.transpose(ptcls.pos), charges=ptcls.charges, pg=2)
    #     elif params.potential.type == 'Yukawa':
    #         out_fmm = fmm.hfmm3d(eps=1.0e-05, zk=1j / params.lambda_TF, sources=np.transpose(ptcls.pos),
    #                          charges=ptcls.charges, pg=2)
    #
    #     potential_energy = ptcls.charges @ out_fmm.pot.real * 4.0 * np.pi / params.fourpie0
    #     ptcls.acc = - np.transpose(ptcls.charges * out_fmm.grad.real / ptcls.mass) / params.fourpie0
    #
    #     return potential_energy








"""
Module for handling Moliere Potential
"""
import numpy as np
import numba as nb


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params : sarkas.core.Parameters
        Simulation's parameters.

    """
    potential.screening_lengths = np.array(potential.screening_lengths)
    potential.screening_charges = np.array(potential.screening_charges)
    params_len = len(potential.screening_lengths)

    moliere_matrix = np.zeros((2 * params_len + 1, params.num_species, params.num_species))

    for i, q1 in enumerate(params.species_charges):
        for j, q2 in enumerate(params.species_charges):

            moliere_matrix[0, i, j] = q1 * q2 / params.fourpie0
            moliere_matrix[1:params_len + 1, i, j] = potential.screening_charges
            moliere_matrix[params_len + 1:, i, j] = potential.screening_lengths

    potential.matrix = moliere_matrix
    potential.force = moliere_force

    # Force error calculated from eq.(43) in Ref.[1]_
    params.force_error = np.sqrt(2.0 * np.pi / potential.screening_lengths.min()) \
                    * np.exp(- potential.rc / potential.screening_lengths.min())
    # Renormalize
    params.force_error *= params.a_ws ** 2 * np.sqrt(params.total_num_ptcls / params.box_volume)


@nb.njit
def moliere_force(r, pot_matrix):
    """ 
    Calculates the PP force between particles using the Moliere Potential.
    
    Parameters
    ----------
    r : float
        Particles' distance.

    pot_matrix : numpy.ndarray
        Moliere potential parameters. 


    Returns
    -------
    phi : float
        Potential

    fr : float
        Force
    """
    """
    Notes
    -----
    See Wilson et al. PRB 15 2458 (1977) for parameters' details
    pot_matrix[0] = Z_1Z_2e^2/(4 np.pi eps_0)
    pot_matrix[1] = C_1
    pot_matrix[2] = C_2
    pot_matrix[3] = C_3
    pot_matrix[4] = b_1
    pot_matrix[5] = b_2
    pot_matrix[6] = b_3
    """

    U = 0.0
    force = 0.0

    for i in range(int(len(pot_matrix[:-1]) / 2)):
        factor1 = r * pot_matrix[i + 4]
        factor2 = pot_matrix[i + 1] / r
        U += factor2 * np.exp(-factor1)
        force += np.exp(-factor1) * factor2 * (1.0 / r + pot_matrix[i + 1])

    force *= pot_matrix[0]
    U *= pot_matrix[0]

    return U, force



"""
Module for handling Yukawa interaction
"""
import numpy as np
from numba import njit
import math as mt


@njit
def yukawa_force_pppm(r, pot_matrix):
    """ 
    Calculates Potential and Force between two particles when the P3M algorithm is chosen.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : array
        Potential matrix. See setup function above.

    Returns
    -------
    U_s_r : float
        Potential value
                
    fr : float
        Force between two particles calculated using eq.(22) in Ref. [Dharuman2017] .

    References
    ----------
    .. [Dharuman2017]: `Dharuman et al. J. Chem. Phys. 146, 024112 (2017) <https://doi.org/10.1063/1.4973842>`_

    """
    kappa = pot_matrix[1]
    alpha = pot_matrix[2]  # Ewald parameter alpha

    kappa_alpha = kappa / alpha
    alpha_r = alpha * r
    kappa_r = kappa * r
    U_s_r = pot_matrix[0] * (0.5 / r) * (np.exp(kappa_r) * mt.erfc(alpha_r + 0.5 * kappa_alpha)
                                         + np.exp(-kappa_r) * mt.erfc(alpha_r - 0.5 * kappa_alpha))
    # Derivative of the exponential term and 1/r
    f1 = (0.5 / r) * np.exp(kappa * r) * mt.erfc(alpha_r + 0.5 * kappa_alpha) * (1.0 / r - kappa)
    f2 = (0.5 / r) * np.exp(-kappa * r) * mt.erfc(alpha_r - 0.5 * kappa_alpha) * (1.0 / r + kappa)
    # Derivative of erfc(a r) = 2a/sqrt(pi) e^{-a^2 r^2}* (x/r)
    f3 = (alpha / np.sqrt(np.pi) / r) * (np.exp(-(alpha_r + 0.5 * kappa_alpha) ** 2) * np.exp(kappa_r)
                                              + np.exp(-(alpha_r - 0.5 * kappa_alpha) ** 2) * np.exp(-kappa_r))
    fr = pot_matrix[0] * (f1 + f2 + f3)

    return U_s_r, fr


@njit
def yukawa_force(r, pot_matrix):
    """ 
    Calculates Potential and Force between two particles.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : array
        It contains potential dependent variables.

    Returns
    -------
    U : float
        Potential.
                
    force : float
        Force between two particles.
    
    """
    U = pot_matrix[0] * np.exp(-pot_matrix[1] * r) / r
    force = U * (1 / r + pot_matrix[1])

    return U, force


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params: sarkas.core.Parameters
        Simulation's parameters.
    """

    if potential.method == "P3M":
        potential.matrix = np.zeros((3, params.num_species, params.num_species))
    else:
        potential.matrix = np.zeros((2, params.num_species, params.num_species))

    twopi = 2.0 * np.pi
    potential.matrix[1, :, :] = 1.0 / params.lambda_TF

    for i, q1 in enumerate(params.species_charges):
        for j, q2 in enumerate(params.species_charges):
            potential.matrix[0, i, j] = q1 * q2 / params.fourpie0

    if potential.method == "PP":
        potential.force = yukawa_force
        # Force error calculated from eq.(43) in Ref.[1]_
        params.force_error = np.sqrt( twopi / params.lambda_TF) * np.exp(- potential.rc / params.lambda_TF)
        # Renormalize
        params.force_error *= params.a_ws ** 2 * np.sqrt(params.total_num_ptcls / params.box_volume)
    elif potential.method == "P3M":
        potential.force = yukawa_force_pppm
        potential.matrix[2, :, :] = potential.pppm_alpha_ewald
        # PP force error calculation. Note that the equation was derived for a single component plasma.
        kappa_over_alpha = - 0.25 * (potential.matrix[1, 0, 0] / potential.matrix[2, 0, 0]) ** 2
        alpha_times_rcut = - (potential.matrix[2, 0, 0] * potential.rc) ** 2
        params.pppm_pp_err = 2.0 * np.exp(kappa_over_alpha + alpha_times_rcut) / np.sqrt(potential.rc)
        params.pppm_pp_err *= np.sqrt(params.total_num_ptcls) * params.a_ws ** 2 / np.sqrt(params.box_volume)



from IPython import get_ipython

if get_ipython().__class__.__name__ == 'ZMQInteractiveShell':
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# Sarkas Modules
import sarkas.tools.observables as obs


class TransportCoefficient:
    """
    Transport Coefficients class.
    """

    def __init__(self):
        pass

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = 'Transport( \n'
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ')'
        return disp

    @staticmethod
    def pretty_print(observable, tc_name):
        """Print to screen the location where data is stored and other relevant information.

        Parameters
        ----------

        observable: sarkas.tools.observables.Observable
            Physical quantity of the ACF.

        tc_name: str
            Name of Transport coefficient to calculate.
        """

        print('Data saved in: \n', os.path.join(observable.saving_dir, tc_name + '_' + observable.job_id + '.h5'))
        print('\nNo. of slices = {}'.format(observable.no_slices))
        print('No. dumps per slice = {}'.format(int(observable.slice_steps / observable.dump_step)))

        print('Time interval of autocorrelation function = {:.4e} [s] ~ {} w_p T'.format(
            observable.dt * observable.slice_steps,
            int(observable.dt * observable.slice_steps * observable.total_plasma_frequency)))

    @staticmethod
    def electrical_conductivity(params,
                                phase: str = 'production',
                                compute_acf: bool = True,
                                no_slices: int = 1,
                                plot: bool = True,
                                show: bool = False,
                                figname: str = None,
                                **kwargs):
        """
        Calculate electrical conductivity from current auto-correlation function.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters.

        phase : str
            Phase to analyze. Default = 'production'.

        show : bool
            Flag for prompting plot to screen.

        Returns
        -------
        coefficient : pandas.DataFrame
            Pandas dataframe containing the value of the transport coefficient as a function of integration time.

        """

        print('\n\n{:=^70} \n'.format(' Electrical Conductivity '))
        coefficient = pd.DataFrame()

        if compute_acf:
            jc_acf = obs.ElectricCurrent()
            jc_acf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            jc_acf.compute()

        else:
            jc_acf = obs.ElectricCurrent()
            jc_acf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            jc_acf.parse()

        # Print some info
        TransportCoefficient.pretty_print(jc_acf, 'ElectricalConductivity')

        no_int = jc_acf.slice_steps

        # to_numpy creates a 2d-array, hence the [:,0]
        time = jc_acf.dataframe[("Time")].to_numpy()[:, 0]

        coefficient["Time"] = time
        jc_str = "Electric Current ACF"
        sigma_str = "Electrical Conductivity"
        for isl in tqdm(range(jc_acf.no_slices), disable = not jc_acf.verbose):
            sigma_ij = np.zeros(jc_acf.slice_steps)

            integrand = np.array(jc_acf.dataframe[(jc_str, "Total", "slice {}".format(isl))])

            for it in range(1, no_int):
                sigma_ij[it] = np.trapz(integrand[:it] / integrand[0], x=time[:it])

                coefficient[sigma_str + "_slice {}".format(isl)] = sigma_ij[:]

        col_str = [sigma_str + "_slice {}".format(isl) for isl in range(jc_acf.no_slices)]
        coefficient[sigma_str + "_Mean"] = coefficient[col_str].mean(axis=1)
        coefficient[sigma_str + "_Std"] = coefficient[col_str].std(axis=1)

        coefficient.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in coefficient.columns])
        coefficient.to_hdf(
            os.path.join(jc_acf.saving_dir, 'ElectricalConductivity_' + jc_acf.job_id + '.h5'),
            mode='w',
            key='conductivity',
            index=False)

        if plot or figname:
            # Make the plot
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 7))
            ax3 = ax1.twiny()
            ax4 = ax2.twiny()
            # extra space for the second axis at the bottom
            # fig.subplots_adjust(bottom=0.2)

            acf_avg = jc_acf.dataframe[(jc_str, "Total", "Mean")]
            acf_std = jc_acf.dataframe[(jc_str, "Total", "Std")]

            d_avg = coefficient[(sigma_str, "Mean")]
            d_std = coefficient[(sigma_str, "Std")]

            # Calculate axis multipliers and labels
            xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, d_avg, "Time", "Conductivity", jc_acf.units)

            # ACF
            ax1.plot(xmul * time, acf_avg / acf_avg.iloc[0], label=r'Current $J$')
            ax1.fill_between(
                xmul * time,
                (acf_avg + acf_std) / (acf_avg.iloc[0] + acf_std.iloc[0]),
                (acf_avg - acf_std) / (acf_avg.iloc[0] - acf_std.iloc[0]), alpha=0.2)

            # Coefficient
            ax2.plot(xmul * time, ymul * d_avg, label=r'$\sigma$')
            ax2.fill_between(xmul * time, ymul * (d_avg + d_std), ymul * (d_avg - d_std), alpha=0.2)

            xlims = (xmul * time[1], xmul * time[-1] * 1.5)

            ax1.set(xlim=xlims, xscale='log', ylabel=r'Electric Current ACF', xlabel=r"Time difference" + xlbl)
            ax2.set(xlim=xlims, xscale='log', ylabel=r'Conductivity' + ylbl, xlabel=r"$\tau$" + xlbl)

            ax1.legend(loc='best')
            ax2.legend(loc='best')
            # Finish the index axes
            for axi in [ax3, ax4]:
                axi.grid(alpha=0.1)
                axi.set(xlim=(1, jc_acf.slice_steps * 1.5), xscale='log', xlabel='Index')

            fig.tight_layout()
            if figname:
                fig.savefig(os.path.join(jc_acf.saving_dir, figname))
            else:
                fig.savefig(os.path.join(jc_acf.saving_dir, 'Plot_ElectricConductivity_' + jc_acf.job_id + '.png'))

            if show:
                fig.show()

        return coefficient

    @staticmethod
    def diffusion(params,
                  phase: str = 'production',
                  compute_acf: bool = True,
                  no_slices: int = 1,
                  plot: bool = True,
                  show: bool = False,
                  figname: str = None,
                  **kwargs):
        """
        Calculate the self-diffusion coefficient from the velocity auto-correlation function.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters.

        phase : str, optional
            Phase to analyze. Default = 'production'.

        compute_acf : bool, optional
            Flag for recalculating the ACF. Default = True.
            If False it will read in the data from the dataframe.

        no_slices : int, optional
            Number of slices of the simulation. Default = 1.

        plot : bool, optional
            Flag to plot transport coefficient with corresponding autocorrelation function. Default = True.

        show : bool, optional
            Flag for prompting plot to screen.

        figname : str, optional
            Name with which to save the file. It automatically saves it in the correct directory.

        **kwargs:
            Arguments to pass :meth:`sarkas.tools.observables.VelocityAutoCorrelationFunction`

        Returns
        -------
        coefficient : pandas.DataFrame
            Pandas dataframe containing the value of the transport coefficient as a function of integration time.

        """
        print('\n\n{:=^70} \n'.format(' Diffusion Coefficient '))
        coefficient = pd.DataFrame()
        if compute_acf:
            vacf = obs.VelocityAutoCorrelationFunction()
            vacf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            vacf.compute()

        else:
            vacf = obs.VelocityAutoCorrelationFunction()
            vacf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            vacf.parse()

        TransportCoefficient.pretty_print(vacf, 'Diffusion')
        time = vacf.dataframe["Time"].to_numpy()[:, 0]
        coefficient["Time"] = time
        vacf_str = 'VACF'
        const = 1.0 / 3.0
        if not params.magnetized:
            # Loop over time slices
            for isl in tqdm(range(vacf.no_slices), disable = not vacf.verbose):
                # Initialize the temporary diffusion container
                D = np.zeros((params.num_species, vacf.slice_steps))
                # Iterate over the number of species
                for i, sp in enumerate(params.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    # Grab vacf data of each slice
                    integrand = np.array(vacf.dataframe[(sp_vacf_str, "Total", "slice {}".format(isl))])
                    # Integrate each timestep
                    for it in range(1, len(time)):
                        D[i, it] = const * np.trapz(integrand[:it], x=time[:it])

                    coefficient["{} Diffusion_slice {}".format(sp, isl)] = D[i, :]

            # Average and std of each diffusion coefficient.
            for isp, sp in enumerate(params.species_names):
                col_str = ["{} Diffusion_slice {}".format(sp, isl) for isl in range(vacf.no_slices)]

                coefficient["{} Diffusion_Mean".format(sp)] = coefficient[col_str].mean(axis=1)
                coefficient["{} Diffusion_Std".format(sp)] = coefficient[col_str].std(axis=1)

        else:
            # Loop over time slices
            for isl in tqdm(range(vacf.no_slices), disable = not vacf.verbose):
                # Initialize the temporary diffusion container
                D = np.zeros((params.num_species, 2, len(time)))
                # Iterate over the number of species
                for i, sp in enumerate(params.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    integrand_par = np.array(vacf.dataframe[(sp_vacf_str, 'Z', "slice {}".format(isl))])
                    integrand_perp = np.array(vacf.dataframe[(sp_vacf_str, 'X', "slice {}".format(isl))]) + \
                                     np.array(vacf.dataframe[(sp_vacf_str, 'Y', "slice {}".format(isl))])

                    for it in range(1, len(time)):
                        D[i, 0, it] = np.trapz(integrand_par[:it], x=time[:it])
                        D[i, 1, it] = 0.5 * np.trapz(integrand_perp[:it], x=time[:it])

                    coefficient["{} Parallel Diffusion_slice {}".format(sp, isl)] = D[i, 0, :]
                    coefficient["{} Perpendicular Diffusion_slice {}".format(sp, isl)] = D[i, 1, :]

            # Add the average and std of perp and par VACF to its dataframe
            for isp, sp in enumerate(params.species_names):
                par_col_str = ["{} Z Velocity ACF slice {}".format(sp, isl) for isl in range(vacf.no_slices)]

                vacf.dataframe["{} Parallel Velocity ACF avg".format(sp)] = vacf.dataframe[par_col_str].mean(axis=1)
                vacf.dataframe["{} Parallel Velocity ACF std".format(sp)] = vacf.dataframe[par_col_str].std(axis=1)

                x_col_str = ["{} X Velocity ACF slice {}".format(sp, isl) for isl in range(vacf.no_slices)]
                y_col_str = ["{} Y Velocity ACF slice {}".format(sp, isl) for isl in range(vacf.no_slices)]

                perp_vacf = 0.5 * (np.array(vacf.dataframe[x_col_str]) + np.array(vacf.dataframe[y_col_str]))
                vacf.dataframe["{} Perpendicular Velocity ACF avg".format(sp)] = perp_vacf.mean(axis=1)
                vacf.dataframe["{} Perpendicular Velocity ACF std".format(sp)] = perp_vacf.std(axis=1)

                # Average and std of each diffusion coefficient.
                par_col_str = ["{} Parallel Diffusion slice {}".format(sp, isl) for isl in range(vacf.no_slices)]
                perp_col_str = ["{} Perpendicular Diffusion slice {}".format(sp, isl) for isl in range(vacf.no_slices)]

                coefficient["{} Parallel Diffusion avg".format(sp)] = coefficient[par_col_str].mean(axis=1)
                coefficient["{} Parallel Diffusion std".format(sp)] = coefficient[par_col_str].std(axis=1)

                coefficient["{} Perpendicular Diffusion avg".format(sp)] = coefficient[perp_col_str].mean(axis=1)
                coefficient["{} Perpendicular Diffusion std".format(sp)] = coefficient[perp_col_str].std(axis=1)

            # Save the updated dataframe
            vacf.dataframe.to_csv(vacf.filename_csv, index=False, encoding='utf-8')

        # Endif magnetized.
        coefficient.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in coefficient.columns])
        # Save the coefficient's data
        coefficient.to_hdf(
            os.path.join(vacf.saving_dir, 'Diffusion_' + vacf.job_id + '.h5'),
            mode='w',
            key='diffusion')

        if plot or figname:
            # Make the plot
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 7))
            # index axes
            ax3 = ax1.twiny()
            ax4 = ax2.twiny()
            # extra space for the second axis at the bottom
            fig.subplots_adjust(bottom=0.2)

            if not params.magnetized:
                for isp, sp in enumerate(params.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    acf_avg = vacf.dataframe[(sp_vacf_str, "Total", "Mean")]
                    acf_std = vacf.dataframe[(sp_vacf_str, "Total", "Std")]

                    d_avg = coefficient[("{} Diffusion".format(sp), "Mean")]
                    d_std = coefficient[("{} Diffusion".format(sp), "Std")]

                    # Calculate axis multipliers and labels
                    xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, d_avg, "Time", "Diffusion", vacf.units)

                    ax1.plot(xmul * time, acf_avg / acf_avg[0], label=r'$D_{' + sp + '}$')
                    ax1.fill_between(
                        xmul * time,
                        (acf_avg + acf_std) / (acf_avg[0] + acf_std[0]),
                        (acf_avg - acf_std) / (acf_avg[0] - acf_std[0]), alpha=0.2)

                    ax2.plot(xmul * time, ymul * d_avg, label=r'$D_{' + sp + '}$')
                    ax2.fill_between(xmul * time, ymul * (d_avg + d_std), ymul * (d_avg - d_std), alpha=0.2)
            else:
                for isp, sp in enumerate(params.species_names):
                    par_acf_avg = vacf.dataframe["{} Parallel Velocity ACF avg".format(sp)]
                    par_acf_std = vacf.dataframe["{} Parallel Velocity ACF std".format(sp)]

                    par_d_avg = coefficient["{} Parallel Diffusion avg".format(sp)]
                    par_d_std = coefficient["{} Parallel Diffusion std".format(sp)]

                    perp_acf_avg = vacf.dataframe["{} Perpendicular Velocity ACF avg".format(sp)]
                    perp_acf_std = vacf.dataframe["{} Perpendicular Velocity ACF std".format(sp)]

                    perp_d_avg = coefficient["{} Perpendicular Diffusion avg".format(sp)]
                    perp_d_std = coefficient["{} Perpendicular Diffusion std".format(sp)]

                    # Calculate axis multipliers and labels
                    xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, perp_d_avg, "Time", "Diffusion", vacf.units)
                    # ACF Parallel plot
                    ax1.plot(xmul * time, par_acf_avg / par_acf_avg.iloc[0], label=sp + r' $Z_{\parallel}$')
                    ax1.fill_between(
                        xmul * time,
                        (par_acf_avg + par_acf_std) / (par_acf_avg.iloc[0] + par_acf_std.iloc[0]),
                        (par_acf_avg - par_acf_std) / (par_acf_avg.iloc[0] - par_acf_std.iloc[0]), alpha=0.2)
                    # Coefficient Parallel plot
                    ax2.plot(xmul * time, ymul * par_d_avg, label=sp + r' $D_{\parallel}$')
                    ax2.fill_between(xmul * time, ymul * (par_d_avg + par_d_std), ymul * (par_d_avg - par_d_std),
                                     alpha=0.2)
                    # ACF Perpendicular Plot
                    ax1.plot(xmul * time, perp_acf_avg / perp_acf_avg.iloc[0], label=sp + r' $Z_{\perp}$')
                    ax1.fill_between(
                        xmul * time,
                        (perp_acf_avg + perp_acf_std) / (perp_acf_avg.iloc[0] + perp_acf_std.iloc[0]),
                        (perp_acf_avg - perp_acf_std) / (perp_acf_avg.iloc[0] - perp_acf_std.iloc[0]), alpha=0.2)
                    # Coefficient Perpendicular Plot
                    ax2.plot(xmul * time, ymul * perp_d_avg, label=sp + r' $D_{\perp}$')
                    ax2.fill_between(xmul * time, ymul * (perp_d_avg + perp_d_std), ymul * (perp_d_avg - perp_d_std),
                                     alpha=0.2)

            xlims = (xmul * time[1], xmul * time[-1] * 1.5)

            ax1.set(xlim=xlims, xscale='log', ylabel=r'Velocity ACF', xlabel=r"Time difference" + xlbl)
            ax2.set(xlim=xlims, xscale='log', ylabel=r'Diffusion' + ylbl, xlabel=r"$\tau$" + xlbl)

            ax1.legend(loc='best')
            ax2.legend(loc='best')

            # Finish the index axes
            for axi in [ax3, ax4]:
                axi.grid(alpha=0.1)
                axi.set(xlim=[1, vacf.slice_steps * 1.5], xscale='log', xlabel='Index')

            fig.tight_layout()
            if figname:
                fig.savefig(os.path.join(vacf.saving_dir, figname))
            else:
                fig.savefig(os.path.join(vacf.saving_dir, 'Plot_Diffusion_' + vacf.job_id + '.png'))

            if show:
                fig.show()

        return coefficient

    @staticmethod
    def interdiffusion(params,
                       phase: str = 'production',
                       compute_acf: bool = True,
                       no_slices: int = 1,
                       plot: bool = True,
                       show: bool = False,
                       figname: str = None,
                       **kwargs):
        """
        Calculate the interdiffusion coefficients from the diffusion flux auto-correlation function.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters.

        phase : str, optional
            Phase to compute. Default = 'production'.

        compute_acf : bool
            Flag for recalculating the ACF. Default = True.
            If False it will read in the data from the dataframe.

        no_slices : int, optional
            Number of slices of the simulation. Default = 1.

        plot : bool, optional
            Flag to plot transport coefficient with corresponding autocorrelation function. Default = True.

        show : bool, optional
            Flag for prompting plot to screen when using IPython kernel. Default = False.

        figname : str, optional
            Name with which to save the file. It automatically saves it in the correct directory.

        **kwargs:
            Arguments to pass :meth:`sarkas.tools.observables.DiffusionFlux`

        Returns
        -------
        coefficient : pandas.DataFrame
            Pandas dataframe containing the value of the transport coefficient as a function of integration time.

        """
        print('\n\n{:=^70} \n'.format(' Interdiffusion Coefficient '))
        coefficient = pd.DataFrame()
        if compute_acf:
            jc_acf = obs.DiffusionFlux()
            jc_acf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            jc_acf.compute()

        else:
            jc_acf = obs.DiffusionFlux()
            jc_acf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            jc_acf.parse()

        # Print some info
        TransportCoefficient.pretty_print(jc_acf, 'Interdiffusion')
        no_int = jc_acf.slice_steps
        no_fluxes_acf = jc_acf.no_fluxes_acf
        # Normalization constant
        const = 1. / (3.0 * jc_acf.total_num_ptcls * jc_acf.species_concentrations.prod())

        # to_numpy creates a 2d-array, hence the [:,0]
        time = jc_acf.dataframe[("Time")].to_numpy()[:, 0]

        coefficient["Time"] = time
        df_str = "Diffusion Flux ACF"
        id_str = "Inter Diffusion Flux"
        for isl in tqdm(range(jc_acf.no_slices), disable= not jc_acf.verbose):
            D_ij = np.zeros((no_fluxes_acf, jc_acf.slice_steps))

            for ij in range(no_fluxes_acf):
                integrand = np.array(jc_acf.dataframe[(df_str + " {}".format(ij), "Total", "slice {}".format(isl))])

                for it in range(1, no_int):
                    D_ij[ij, it] = const * np.trapz(integrand[:it], x=time[:it])

                coefficient[id_str + " {}_slice {}".format(ij, isl)] = D_ij[ij, :]

        # Average and Std of slices
        for ij in range(no_fluxes_acf):
            col_str = [id_str + " {}_slice {}".format(ij, isl) for isl in range(jc_acf.no_slices)]
            coefficient[id_str + " {}_Mean".format(ij)] = coefficient[col_str].mean(axis=1)
            coefficient[id_str + " {}_Std".format(ij)] = coefficient[col_str].std(axis=1)

        coefficient.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in coefficient.columns])
        coefficient.to_hdf(
            os.path.join(jc_acf.saving_dir, 'InterDiffusion_' + jc_acf.job_id + '.h5'),
            mode='w',
            key='interdiffusion',
            index=False)

        if plot or figname:
            # Make the plot
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 7))
            ax3 = ax1.twiny()
            ax4 = ax2.twiny()
            # extra space for the second axis at the bottom
            # fig.subplots_adjust(bottom=0.2)
            for flux in range(jc_acf.no_fluxes):
                acf_avg = jc_acf.dataframe[(df_str + " {}".format(flux), "Total", "Mean")]
                acf_std = jc_acf.dataframe[(df_str + " {}".format(flux), "Total", "Std")]

                d_avg = coefficient[(id_str + " {}".format(flux), "Mean")]
                d_std = coefficient[(id_str + " {}".format(flux), "Std")]

                # Calculate axis multipliers and labels
                xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, d_avg, "Time", "Diffusion", jc_acf.units)

                # ACF
                ax1.plot(xmul * time, acf_avg / acf_avg.iloc[0], label=r'Flux $J_{}$'.format(flux))
                ax1.fill_between(
                    xmul * time,
                    (acf_avg + acf_std) / (acf_avg.iloc[0] + acf_std.iloc[0]),
                    (acf_avg - acf_std) / (acf_avg.iloc[0] - acf_std.iloc[0]), alpha=0.2)

                # Coefficient
                ax2.plot(xmul * time, ymul * d_avg, label=r'IDF $D_{}$'.format(flux))
                ax2.fill_between(xmul * time, ymul * (d_avg + d_std), ymul * (d_avg - d_std), alpha=0.2)

            xlims = (xmul * time[1], xmul * time[-1] * 1.5)

            ax1.set(xlim=xlims, xscale='log', ylabel=r'Diffusion Flux ACF', xlabel=r"Time difference" + xlbl)
            ax2.set(xlim=xlims, xscale='log', ylabel=r'Inter Diffusion' + ylbl, xlabel=r"$\tau$" + xlbl)

            ax1.legend(loc='best')
            ax2.legend(loc='best')
            # Finish the index axes
            for axi in [ax3, ax4]:
                axi.grid(alpha=0.1)
                axi.set(xlim=(1, jc_acf.slice_steps * 1.5), xscale='log', xlabel='Index')

            fig.tight_layout()
            if figname:
                fig.savefig(os.path.join(jc_acf.saving_dir, figname))
            else:
                fig.savefig(os.path.join(jc_acf.saving_dir, 'Plot_InterDiffusion_' + jc_acf.job_id + '.png'))

            if show:
                fig.show()

        # Save to dataframe

        return coefficient

    @staticmethod
    def viscosity(params,
                  phase: str = 'production',
                  compute_acf: bool = True,
                  no_slices: int = 1,
                  plot: bool = True,
                  plot_quantities: list = ["Bulk Viscosity"],
                  show: bool = False,
                  figname: str = None,
                  **kwargs
                  ):
        """
        Calculate bulk and shear viscosity from pressure auto-correlation function.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters.

        phase : str
            Phase to analyze. Default = 'production'.

        show : bool
            Flag for prompting plot to screen.

        Returns
        -------
        coefficient : pandas.DataFrame
            Pandas dataframe containing the value of the transport coefficient as a function of integration time.

        """
        print('\n\n{:=^70} \n'.format(' Viscosity Coefficient '))
        coefficient = pd.DataFrame()

        energies = obs.Thermodynamics()
        energies.setup(params, phase)
        energies.parse()

        if compute_acf:
            pt = obs.PressureTensor()
            pt.setup(params, phase, no_slices, **kwargs)
            pt.compute()
        else:
            pt = obs.PressureTensor()
            pt.setup(params, phase, no_slices, **kwargs)
            pt.parse()

        TransportCoefficient.pretty_print(pt, 'Viscosities')

        time = pt.dataframe["Time"].to_numpy()[:, 0]
        coefficient["Time"] = time
        dim_lbl = ['x', 'y', 'z']

        pt_str = "Pressure Tensor ACF"
        start_steps = 0
        end_steps = 0

        for isl in tqdm(range(pt.no_slices), disable=not pt.verbose):
            end_steps += pt.slice_steps

            beta = (np.mean(energies.dataframe["Temperature"].iloc[start_steps:end_steps]) * pt.kB) ** (-1.0)
            const = params.box_volume * beta
            # Calculate Bulk Viscosity
            bulk_viscosity = np.zeros(pt.slice_steps)
            # Bulk viscosity is calculated from the fluctuations of the pressure eq. 2.124a Allen & Tilsdeley
            bulk_integrand = np.copy(pt.dataframe[("Delta Pressure ACF", "slice {}".format(isl))])

            for it in range(1, pt.slice_steps):
                bulk_viscosity[it] = const * np.trapz(bulk_integrand[:it], x=time[:it])

            coefficient["Bulk Viscosity_slice {}".format(isl)] = bulk_viscosity

            shear_viscosity = np.zeros((params.dimensions, params.dimensions, pt.slice_steps))

            for i, ax1 in enumerate(dim_lbl):
                for j, ax2 in enumerate(dim_lbl):

                    integrand = pt.dataframe[(pt_str + " {}{}".format(ax1, ax2), 'slice {}'.format(isl))].to_numpy()
                    for it in range(1, pt.slice_steps):
                        shear_viscosity[i, j, it] = const * np.trapz(integrand[:it], x=time[:it])

                    coefficient["Shear Viscosity Tensor {}{}_slice {}".format(ax1, ax2, isl)] = shear_viscosity[i, j, :]
            start_steps += pt.slice_steps

        # Now average the slice
        col_str = ["Bulk Viscosity_slice {}".format(isl) for isl in range(pt.no_slices)]
        coefficient["Bulk Viscosity_Mean"] = coefficient[col_str].mean(axis=1)
        coefficient["Bulk Viscosity_Std"] = coefficient[col_str].std(axis=1)

        sv_str = "Shear Viscosity Tensor"
        for i, ax1 in enumerate(dim_lbl):
            for j, ax2 in enumerate(dim_lbl):
                col_str = [sv_str + " {}{}_slice {}".format(ax1, ax2, isl) for isl in range(pt.no_slices)]
                coefficient[sv_str + " {}{}_Mean".format(ax1, ax2)] = coefficient[col_str].mean(axis=1)
                coefficient[sv_str + " {}{}_Std".format(ax1, ax2)] = coefficient[col_str].std(axis=1)

        coefficient.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in coefficient.columns])
        coefficient.to_hdf(
            os.path.join(pt.saving_dir, 'Viscosities_' + pt.job_id + '.h5'),
            mode='w',
            key='viscosities',
            index=False)

        if plot or figname:
            # Make the plot
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 7))
            ax3 = ax1.twiny()
            ax4 = ax2.twiny()
            # extra space for the second axis at the bottom
            # fig.subplots_adjust(bottom=0.2)
            for ipq, pq in enumerate(plot_quantities):
                if pq[:4].lower() == "bulk":
                    acf_str = "Delta Pressure ACF"
                    acf_avg = pt.dataframe[("Delta Pressure ACF", "Mean")]
                    acf_std = pt.dataframe[("Delta Pressure ACF", "Std")]
                else:
                    # The axis are the last two elements in the string
                    acf_str = "Pressure Tensor ACF " + pq[-2:]
                    acf_avg = pt.dataframe[("Pressure Tensor ACF " + pq[-2:], "Mean")]
                    acf_std = pt.dataframe[("Pressure Tensor ACF " + pq[-2:], "Std")]

                d_avg = coefficient[(pq, "Mean")]
                d_std = coefficient[(pq, "Std")]

                # Calculate axis multipliers and labels
                xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, d_avg, "Time", "Viscosity", pt.units)

                # ACF
                ax1.plot(xmul * time, acf_avg / acf_avg.iloc[0], label=acf_str)
                ax1.fill_between(
                    xmul * time,
                    (acf_avg + acf_std) / (acf_avg.iloc[0] + acf_std.iloc[0]),
                    (acf_avg - acf_std) / (acf_avg.iloc[0] - acf_std.iloc[0]), alpha=0.2)

                # Coefficient
                ax2.plot(xmul * time, ymul * d_avg, label=r'$\eta$')
                ax2.fill_between(xmul * time, ymul * (d_avg + d_std), ymul * (d_avg - d_std), alpha=0.2)

                xlims = (xmul * time[1], xmul * time[-1] * 1.5)

                ax1.set(xlim=xlims, xscale='log', ylabel=acf_str, xlabel=r"Time difference" + xlbl)
                ax2.set(xlim=xlims, xscale='log', ylabel=r'Viscosity' + ylbl, xlabel=r"$\tau$" + xlbl)

                ax1.legend(loc='best')
                ax2.legend(loc='best')
                # Finish the index axes
                for axi in [ax3, ax4]:
                    axi.grid(alpha=0.1)
                    axi.set(xlim=(1, pt.slice_steps * 1.5), xscale='log', xlabel='Index')

                fig.tight_layout()
                if figname:
                    fig.savefig(os.path.join(pt.saving_dir, figname))
                else:
                    fig.savefig(os.path.join(pt.saving_dir, 'Plot_' + pq.replace(" ", "") + '_' + pt.job_id + '.png'))

                if show:
                    fig.show()

        return coefficient



import time
from dataclasses import dataclass, field
from typing import Optional


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


@dataclass
class SarkasTimer:
    """
    Timer class modified from https://realpython.com/python-timer/
    """
    _start_time: Optional[float] = field(default=None, init=False, repr=False)

    def start(self):
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError(f"Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter_ns()

    def stop(self) -> float:
        """
        Stop the timer, and report the elapsed time.

        Returns
        -------
        elapsed_time: float
            Elapsed time in nanoseconds.

        """
        if self._start_time is None:
            raise TimerError(f"Timer is not running. Use .start() to start it")

        # Calculate elapsed time
        elapsed_time = time.perf_counter_ns() - self._start_time
        self._start_time = None

        return elapsed_time

    @staticmethod
    def current() -> float:
        """Grab the current time in nanoseconds."""
        return time.perf_counter_ns()

    @staticmethod
    def time_division(tme : float) -> list:
        """
        Divide time into hours, min, sec, msec, microsec (usec), and nanosec.

        Parameters
        ----------
        tme : float
            Time in nanoseconds.

        Returns
        -------
        : list
            [hours, min, sec, msec, microsec (usec), nanosec]

        """
        t_hrs, rem = divmod(tme, 3.6e12)
        t_min, rem_m = divmod(rem, 6e+10)
        t_sec, rem_s = divmod(rem_m, 1e9)
        t_msec, rem_ms = divmod(rem_s, 1e6)
        t_usec, rem_us = divmod(rem_ms, 1e3)
        t_nsec, rem_ns = divmod(rem_us, 1)

        return [t_hrs, t_min, t_sec, t_msec, t_usec, t_nsec]


"""
Module for handling the Quantum Statistical Potential.

Note
----
Notice that in Ref. [Hansen1981]_ the DeBroglie wavelength is defined as

.. math::
   \lambda_{ee} = \dfrac{\hbar}{\sqrt{2 \pi \mu_{ee} k_{B} T} },

while in statistical physics textbooks and Ref. [Glosli2008]_ is defined as

.. math::
   \lambda_{ee} = \dfrac{h}{\sqrt{2 \pi \mu_{ee} k_{B} T} },

References
----------
.. [Hansen1981] `J.P. Hansen and I.R. McDonald, Phys Rev A 23 2041 (1981) <https://doi.org/10.1103/PhysRevA.23.2041>`_
.. [Glosli2008] `J.N. Glosli et al. Phys Rev E 78 025401(R) (2008) <https://doi.org/10.1103/PhysRevE.78.025401>`_
"""
import numpy as np
from numba import njit
import math as mt


def update_params(potential, params):
    """
    Create potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params : object
        Simulation's parameters

    """
    """
    Dev Notes
    for more info and a description of the potential's parameters.
    QSP_matrix[0,:,:] = qi*qj/4*pi*eps0
    QSP_matrix[1,:,:] = 2pi/deBroglie
    QSP_matrix[2,:,:] = e-e Pauli term factor
    QSP_matrix[3,:,:] = e-e Pauli term exponent term
    QSP_matrix[4,:,:] = Ewald parameter
    """
    # Do a bunch of checks
    # P3M algorithm only
    assert potential.method == "P3M", 'QSP interaction can only be calculated using P3M algorithm.'

    # Check for neutrality
    assert params.total_net_charge == 0, 'Total net charge is not zero.'

    # Default attributes
    if not hasattr(potential, 'qsp_type'):
        potential.qsp_type = 'Deutsch'
    if not hasattr(potential, 'qsp_pauli'):
        potential.qsp_pauli = True

    two_pi = 2.0 * np.pi
    four_pi = 2.0 * two_pi
    log2 = np.log(2.0)

    # Redefine ion temperatures and ion total number density
    mask = params.species_names != 'e'
    params.total_ion_temperature = params.species_concentrations[mask].transpose() * params.species_temperatures[mask]
    params.ni = np.sum(params.species_num_dens[mask])

    # Calculate the total and ion Wigner-Seitz Radius from the total density
    params.ai = (3.0 / (four_pi * params.ni)) ** (1.0 / 3.0)  # Ion WS

    deBroglie_const = two_pi * params.hbar2 / params.kB

    QSP_matrix = np.zeros((5, params.num_species, params.num_species))
    for i, name1 in enumerate(params.species_names):
        m1 = params.species_masses[i]
        q1 = params.species_charges[i]

        for j, name2 in enumerate(params.species_names):
            m2 = params.species_masses[j]
            q2 = params.species_charges[j]

            reduced = (m1 * m2) / (m1 + m2)

            if name1 == 'e' or name2 == 'e':
                # Use electron temperature in e-e and e-i interactions
                lambda_deB = np.sqrt(deBroglie_const / (reduced * params.electron_temperature))
            else:
                # Use ion temperature in i-i interactions only
                lambda_deB = np.sqrt(deBroglie_const / (reduced * params.total_ion_temperature))

            if name2 == name1:  # e-e
                QSP_matrix[2, i, j] = log2 * params.kB * params.electron_temperature
                QSP_matrix[3, i, j] = four_pi / (log2 * lambda_deB ** 2)

            QSP_matrix[0, i, j] = q1 * q2 / params.fourpie0
            QSP_matrix[1, i, j] = two_pi / lambda_deB

    if not potential.qsp_pauli:
        QSP_matrix[2, :, :] = 0.0

    QSP_matrix[4, :, :] = potential.pppm_alpha_ewald
    potential.matrix = QSP_matrix

    if potential.qsp_type.lower() == "deutsch":
        potential.force = deutsch_force
        # Calculate the PP Force error from the e-e diffraction term only.
        params.pppm_pp_err = np.sqrt(two_pi * potential.matrix[1, 0, 0])
        params.pppm_pp_err *= np.exp(- potential.rc * potential.matrix[1, 0, 0])
        params.pppm_pp_err *= params.a_ws ** 2 * np.sqrt(params.total_num_ptcls / params.box_volume)

    elif potential.qsp_type.lower() == "kelbg":
        potential.force = kelbg_force
        # TODO: Calculate the PP Force error from the e-e diffraction term only.
        params.pppm_pp_err = np.sqrt(two_pi * potential.matrix[1, 0, 0])
        params.pppm_pp_err *= np.exp(- potential.rc * potential.matrix[1, 0, 0])


@njit
def deutsch_force(r, pot_matrix):
    """ 
    Calculate Deutsch QSP Force between two particles.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : array
        It contains potential dependent variables.
        pot_matrix[0,:,:] = qi*qj/4*pi*eps0
        pot_matrix[1,:,:] = 2pi/deBroglie
        pot_matrix[2,:,:] = e-e Pauli term factor
        pot_matrix[3,:,:] = e-e Pauli term exponent term
        pot_matrix[4,:,:] = Ewald parameter

    Returns
    -------
    U : float
        Potential.
                
    force : float
        Force between two particles.

    """

    A = pot_matrix[0]
    C = pot_matrix[1]
    D = pot_matrix[2]
    F = pot_matrix[3]
    alpha = pot_matrix[4]

    a2 = alpha * alpha
    r2 = r * r

    # Ewald short-range potential and force terms
    U_ewald = A * mt.erfc(alpha * r) / r
    f_ewald = U_ewald / r  # 1/r derivative
    f_ewald += A * (2.0 * alpha / np.sqrt(np.pi) ) * np.exp(- a2 * r2)/r  # erfc derivative

    # Diffraction potential and force term
    U_diff = -A * np.exp(-C * r) / r
    f_diff = U_diff / r  # 1/r derivative
    f_diff += -A * C * np.exp(-C * r) / r  # exp derivative

    # Pauli potential and force terms
    U_pauli = D * np.exp(-F * r2)
    f_pauli = 2.0 * r * D * F * np.exp(-F * r2)

    U = U_ewald + U_diff + U_pauli
    force = f_ewald + f_diff + f_pauli

    return U, force


@njit
def kelbg_force(r, pot_matrix):
    """ 
    Calculates the QSP Force between two particles when the P3M algorithm is chosen.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : array
        It contains potential dependent parameters.
        pot_matrix[0] = qi*qj/4*pi*eps0
        pot_matrix[1] = 2pi/deBroglie
        pot_matrix[2] = e-e Pauli term factor
        pot_matrix[3] = e-e Pauli term exponent term
        pot_matrix[4] = Ewald parameter

    Returns
    -------
    U : float
        Potential.
                
    force : float
        Force between two particles.
    
    """

    A = pot_matrix[0]
    C = pot_matrix[1]
    D = pot_matrix[2]
    F = pot_matrix[3]
    alpha = pot_matrix[4]
    C2 = C * C
    a2 = alpha * alpha
    r2 = r * r

    # Ewald short-range potential and force terms
    U_ewald = A * mt.erfc(alpha * r) / r
    f_ewald = U_ewald / r2  # 1/r derivative
    f_ewald += A * (2.0 * alpha / np.sqrt(np.pi) / r2) * np.exp(- a2 * r2)  # erfc derivative

    # Diffraction potential and force term
    U_diff = -A * np.exp(-C2 * r2 / np.pi) / r
    U_diff += A * C * mt.erfc(C * r / np.sqrt(np.pi))

    f_diff = -A * (2.0 * C2 * r2 + np.pi) * np.exp(- C2 * r2 / np.pi) / (np.pi * r * r2)  # exp(r)/r derivative
    f_diff += 2.0 * A * C2 * np.exp(- C2 * r2 / np.pi) / r / np.pi  # erfc derivative

    # Pauli Term
    U_pauli = D * np.exp(-F * r2)
    f_pauli = 2.0 * D * F * np.exp(-F * r2)

    U = U_ewald + U_diff + U_pauli
    force = f_ewald + f_diff + f_pauli

    return U, force



"""
Module containing the three basic classes: Parameters, Particles, Species.
"""
import numpy as np
import os.path
import sys
import scipy.constants as const


class Parameters:
    """
    Class containing all the constants and physical constants of the simulation.

    Parameters
    ----------
    dic : dict, optional
        Dictionary to be copied.

    Attributes
    ----------
    a_ws : float
        Wigner-Seitz radius. Calculated from the ``total_num_density`` .

    boundary_conditions : str
        Type of boundary conditions.

    box_volume : float
        Volume of simulation box.

    dimensions : int
        Number of non-zero dimensions. Default = 3.

    fourpie0: float
        Electrostatic constant :math:`4\pi \epsilon_0`.

    num_species : int
        Number of species.

    kB : float
        Boltzmann constant obtained from ``scipy.constants``.

    hbar : float
        Reduced Planck's constant.

    hbar2 : float
        Square of reduced Planck's constant.

    a0 : float
        Bohr Radius.

    c0 : float
        Speed of light.

    qe : float
        Elementary charge.

    me : float
        Electron mass.

    eps0 : float
        Vacuum electrical permittivity.

    eV2K : float
        Conversion factor from eV to Kelvin obtained from ``scipy.constants``.

    J2erg : float
        Conversion factor from Joules to erg. Needed for cgs units.

    QFactor : float
        Charge Factor defined as :math:`\mathcal Q = \sum_{i}^{N} q_{i}^2` .

    Lx : float
        Box length in the :math:`x` direction.

    Ly : float
        Box length in the :math:`y` direction.

    Lz : float
        Box length in the :math:`z` direction.

    e1 : float
        Unit vector in the :math:`x` direction.

    e2 : float
        Unit vector in the :math:`y` direction.

    e3 : float
        Unit vector in the :math:`z` direction.

    input_file : str
        YAML Input file with all the simulation's parameters.

    T_desired : float
        Target temperature for the equilibration phase.

    species_num : numpy.ndarray
        Number of particles of each species. Shape = (``num_species``)

    species_concentrations : numpy.ndarray
        Concentration of each species. Shape = (``num_species``)

    species_temperatures : numpy.ndarray
        Initial temperature of each species. Shape = (``num_species``)

    species_masses : numpy.ndarray
        Mass of each species. Shape = (``num_species``)

    species_charges : numpy.ndarray
        Charge of each species. Shape = (``num_species``)

    species_names : list
        Name of each species. Len = (``num_species``)

    species_plasma_frequencies : numpy.ndarray
        Plasma Frequency of each species. Shape = (``num_species``)

    species_num_dens : numpy.ndarray
        Number density of each species. Shape = (``num_species``)

    total_ion_temperature : float
        Total initial ion temperature calculated as `` = species_concentration @ species_temperatures``.

    total_net_charge : float
        Total charge in the system.

    total_num_density : float
        Total number density. Calculated from the sum of ``Species.num_density``.

    total_num_ptcls : int
        Total number of particles. Calculated from the sum of ``Species.num``.

    measure : bool
        Flag for production phase.

    verbose : bool
        Flag for screen output.

    simulations_dir : str
        Name of directory where to store simulations.

    job_dir : str
        Directory name of the current job/run

    production_dir : str
        Directory name where to store simulation's files of the production phase. Default = 'Production'.

    equilibration_dir : str
        Directory name where to store simulation's file of the equilibration phase. Default = 'Equilibration'.

    preprocessing_dir : str
        Directory name where to store preprocessing files. Default = "PreProcessing".

    postprocessing_dir : str
        Directory name where to store postprocessing files. Default = "PostProcessing".

    prod_dump_dir : str
        Directory name where to store production phase's simulation's checkpoints. Default = 'dumps'.

    eq_dump_dir : str
        Directory name where to store equilibration phase's simulation's checkpoints. Default = 'dumps'.

    job_id : str
        Appendix of all simulation's files.

    log_file : str
        Filename of the simulation's log.

    np_per_side : numpy.ndarray
        Number of particles per simulation's box side.
        The product of its components should be equal to ``total_num_ptcls``.

    pre_run : bool
        Flag for preprocessing phase.

    """

    def __init__(self, dic: dict = None) -> None:

        self.input_file = None
        #
        self.Lx = 0.0
        self.Ly = 0.0
        self.Lz = 0.0
        self.box_lengths = np.zeros(3)
        self.box_volume = 0.0
        self.dimensions = 3
        self.J2erg = 1.0e+7  # erg/J
        self.eps0 = const.epsilon_0
        self.fourpie0 = 4.0 * np.pi * self.eps0
        self.mp = const.physical_constants["proton mass"][0]
        self.me = const.physical_constants["electron mass"][0]
        self.qe = const.physical_constants["elementary charge"][0]
        self.hbar = const.hbar
        self.hbar2 = self.hbar ** 2
        self.c0 = const.physical_constants["speed of light in vacuum"][0]
        self.eV2K = const.physical_constants["electron volt-kelvin relationship"][0]
        self.eV2J = const.physical_constants["electron volt-joule relationship"][0]
        self.a0 = const.physical_constants["Bohr radius"][0]
        self.kB = const.Boltzmann
        self.kB_eV = const.physical_constants["Boltzmann constant in eV/K"][0]
        self.a_ws = 0.0

        # Control and Timing
        self.boundary_conditions = 'periodic'
        self.job_id = None
        self.job_dir = None
        self.log_file = None
        self.measure = False
        self.magnetized = False
        self.plot_style = None
        self.pre_run = False
        self.simulations_dir = "Simulations"
        self.production_dir = 'Production'
        self.magnetization_dir = 'Magnetization'
        self.equilibration_dir = 'Equilibration'
        self.preprocessing_dir = "PreProcessing"
        self.postprocessing_dir = "PostProcessing"
        self.prod_dump_dir = 'dumps'
        self.eq_dump_dir = 'dumps'
        self.mag_dump_dir = 'dumps'
        self.verbose = True
        self.restart_step = None
        self.np_per_side = None
        self.num_species = 1
        self.species_names = None
        self.species_num = None
        self.species_num_dens = None
        self.species_concentrations = None
        self.species_temperatures = None
        self.species_temperatures_eV = None
        self.species_masses = None
        self.species_charges = None
        self.species_plasma_frequencies = None
        self.species_cyclotron_frequencies = None
        self.species_couplings = None

        self.coupling_constant = 0.0
        self.total_num_density = 0.0
        self.total_num_ptcls = 0
        self.total_plasma_frequency = 0.0
        self.total_debye_length = 0.0
        self.total_mass_density = 0.0
        self.total_ion_temperature = 0.0
        self.T_desired = 0.0
        self.total_net_charge = 0.0
        self.QFactor = 0.0

        self.average_charge = None
        self.average_mass = None
        self.hydrodynamic_frequency = None

        if dic:
            self.from_dict(dic)

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = 'Parameters( \n'
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ')'
        return disp

    def from_dict(self, input_dict: dict):
        """
        Update attributes from input dictionary.

        Parameters
        ----------
        input_dict: dict
            Dictionary to be copied.

        """
        self.__dict__.update(input_dict)

    def setup(self, species):
        """
        Setup simulations' parameters.

        Parameters
        ----------
        species : list
            List of ``sarkas.core.Species`` objects.

        """
        self.check_units()
        self.calc_parameters(species)
        self.calc_coupling_constant(species)

    def check_units(self) -> None:
        """Adjust default physical constants for cgs unit system and check for LJ potential."""
        # Physical constants
        if self.units == "cgs":
            self.kB *= self.J2erg
            self.c0 *= 1e2  # cm/s
            self.mp *= 1e3
            # Coulomb to statCoulomb conversion factor. See https://en.wikipedia.org/wiki/Statcoulomb
            C2statC = 1.0e-01 * self.c0
            self.hbar = self.J2erg * self.hbar
            self.hbar2 = self.hbar ** 2
            self.qe *= C2statC
            self.me *= 1.0e3
            self.eps0 = 1.0
            self.fourpie0 = 1.0
            self.a0 *= 1e2

        if self.potential_type == 'LJ':
            self.fourpie0 = 1.0
            self.species_lj_sigmas = np.zeros(self.num_species)

    def calc_parameters(self, species):
        """
        Assign the parsed parameters.

        Parameters
        ----------
        species : list
            List of ``sarkas.core.Species`` objects.

        """

        self.num_species = len(species)
        # Initialize the arrays containing species attributes. This is needed for postprocessing
        self.species_names = []
        self.species_num = np.zeros(self.num_species, dtype=int)
        self.species_num_dens = np.zeros(self.num_species)
        self.species_concentrations = np.zeros(self.num_species)
        self.species_temperatures = np.zeros(self.num_species)
        self.species_temperatures_eV = np.zeros(self.num_species)
        self.species_masses = np.zeros(self.num_species)
        self.species_charges = np.zeros(self.num_species)
        self.species_plasma_frequencies = np.zeros(self.num_species)
        self.species_cyclotron_frequencies = np.zeros(self.num_species)
        self.species_couplings = np.zeros(self.num_species)

        # Loop over species and assign missing attributes
        # Collect species properties in single arrays
        wp_tot_sq = 0.0
        lambda_D = 0.0

        # Initialization of attributes
        self.total_num_ptcls = 0
        self.total_num_density = 0.0
        # Calculate species attributes and add them to simulation parameters
        for i, sp in enumerate(species):
            self.total_num_ptcls += sp.num

            # Calculate the mass of the species from the atomic weight if given
            if sp.atomic_weight:
                # Choose between atomic mass constant or proton mass
                # u = const.physical_constants["atomic mass constant"][0]
                sp.mass = self.mp * sp.atomic_weight

            # Calculate the mass of the species from the mass density if given
            if sp.mass_density:
                Av = const.physical_constants["Avogadro constant"][0]
                sp.number_density = sp.mass_density * Av / sp.atomic_weight
                self.total_num_density += sp.number_density

            assert sp.number_density, "{} number density not defined".format(sp.name)

            # Update arrays of species information
            self.total_num_density += sp.number_density
            self.species_names.append(sp.name)
            self.species_num[i] = sp.num
            self.species_masses[i] = sp.mass
            # Calculate the temperature in K if eV has been provided and vice versa
            if sp.temperature_eV:
                sp.temperature = self.eV2K * sp.temperature_eV
                self.species_temperatures_eV[i] = sp.temperature_eV
            else:
                # Convert to eV and save
                sp.temperature_eV = sp.temperature / self.eV2K
                self.species_temperatures_eV[i] = sp.temperature_eV

            self.species_temperatures[i] = sp.temperature

            # Calculate the species charge based on the inputs
            if sp.charge:
                self.species_charges[i] = sp.charge
                sp.Z = sp.charge / self.qe
            elif sp.Z:
                self.species_charges[i] = sp.Z * self.qe
                sp.charge = sp.Z * self.qe
            elif sp.epsilon:
                # Lennard-Jones potentials don't have charge but have the equivalent epsilon.
                sp.charge = np.sqrt(sp.epsilon)
                sp.Z = 1.0
                self.species_charges[i] = np.sqrt(sp.epsilon)
            else:
                sp.charge = 0.0
                sp.Z = 0.0
                self.species_charges[i] = 0.0

            # Calculate the (total) plasma frequency, QFactor, debye_length
            if not self.potential_type == "LJ":
                # Q^2 factor see eq.(2.10) in Ballenegger et al. J Chem Phys 128 034109 (2008)
                sp.QFactor = sp.num * sp.charge ** 2
                self.QFactor += sp.QFactor / self.fourpie0

                sp.calc_plasma_frequency(self.fourpie0)
                wp_tot_sq += sp.plasma_frequency ** 2
                sp.calc_debye_length(self.kB, self.fourpie0)
                lambda_D += sp.debye_length ** 2
            else:
                sp.QFactor = 0.0
                self.QFactor += sp.QFactor / self.fourpie0
                constant = 4.0 * np.pi * sp.number_density * sp.sigma ** 2
                sp.calc_plasma_frequency(constant)
                wp_tot_sq += sp.plasma_frequency ** 2
                sp.calc_debye_length(self.kB, constant)
                lambda_D += sp.debye_length ** 2
                self.species_lj_sigmas[i] = sp.sigma

            # Calculate cyclotron frequency in case of a magnetized simulation
            if self.magnetized:

                if self.units == "cgs":
                    sp.calc_cyclotron_frequency(np.linalg.norm(self.magnetic_field) / self.c0)
                else:
                    sp.calc_cyclotron_frequency(np.linalg.norm(self.magnetic_field))

                sp.beta_c = sp.cyclotron_frequency / sp.plasma_frequency
                self.species_cyclotron_frequencies[i] = sp.cyclotron_frequency

            self.species_plasma_frequencies[i] = sp.plasma_frequency
            self.species_num_dens[i] = sp.number_density

        self.total_mass_density = self.species_masses.transpose() @ self.species_num_dens

        # Calculate species concentrations
        for i, sp in enumerate(species):
            sp.concentration = float(sp.num / self.total_num_ptcls)
            self.species_concentrations[i] = float(sp.num / self.total_num_ptcls)
            # Calculate species mass properties
            if sp.mass_density is None:
                sp.mass_density = sp.concentration * self.total_mass_density
                if sp.atomic_weight is None:
                    sp.atomic_weight = sp.mass / self.mp

        # Calculate total quantities
        self.total_net_charge = np.transpose(self.species_charges) @ self.species_num
        self.total_ion_temperature = np.transpose(self.species_concentrations) @ self.species_temperatures
        self.total_plasma_frequency = np.sqrt(wp_tot_sq)
        self.total_debye_length = np.sqrt(lambda_D)

        self.average_charge = np.transpose(self.species_charges) @ self.species_concentrations
        self.average_mass = np.transpose(self.species_masses) @ self.species_concentrations
        # Hydrodynamic Frequency aka Virtual Average Atom
        self.hydrodynamic_frequency = np.sqrt(4.0 * np.pi * self.average_charge ** 2 * self.total_num_density /
                                              (self.fourpie0 * self.average_mass))

        # Simulation Box Parameters
        # Wigner-Seitz radius calculated from the total number density
        self.a_ws = (3.0 / (4.0 * np.pi * self.total_num_density)) ** (1. / 3.)

        # Calculate simulation's box parameters
        if self.np_per_side:
            msg = "Number of particles per dimension does not match total number of particles."
            assert int(np.prod(self.np_per_side)) == self.total_num_ptcls, msg

            self.Lx = self.a_ws * self.np_per_side[0] * (4.0 * np.pi / 3.0) ** (1.0 / 3.0)
            self.Ly = self.a_ws * self.np_per_side[1] * (4.0 * np.pi / 3.0) ** (1.0 / 3.0)
            self.Lz = self.a_ws * self.np_per_side[2] * (4.0 * np.pi / 3.0) ** (1.0 / 3.0)
        else:
            self.Lx = self.a_ws * (4.0 * np.pi * self.total_num_ptcls / 3.0) ** (1.0 / 3.0)
            self.Ly = self.a_ws * (4.0 * np.pi * self.total_num_ptcls / 3.0) ** (1.0 / 3.0)
            self.Lz = self.a_ws * (4.0 * np.pi * self.total_num_ptcls / 3.0) ** (1.0 / 3.0)

        self.box_lengths = np.array([self.Lx, self.Ly, self.Lz])  # box length vector

        # Dev Note: The following are useful for future geometries
        self.e1 = np.array([self.Lx, 0.0, 0.0])
        self.e2 = np.array([0.0, self.Ly, 0.0])
        self.e3 = np.array([0.0, 0.0, self.Lz])

        self.box_volume = abs(np.dot(np.cross(self.e1, self.e2), self.e3))

        self.dimensions = np.count_nonzero(self.box_lengths)  # no. of dimensions
        # Transform the list of species names into a np.array
        self.species_names = np.array(self.species_names)
        # Redundancy!!!
        self.T_desired = self.total_ion_temperature

    def calc_coupling_constant(self, species):
        """
        Calculate the coupling constant of each species and the total coupling constant. For more information see
        the theory pages.

        Parameters
        ----------
        species: list
            List of ``sarkas.core.Species`` objects.

        """
        z_avg = np.transpose(self.species_charges) @ self.species_concentrations

        for i, sp in enumerate(species):
            const = self.fourpie0 * self.kB
            sp.calc_coupling(self.a_ws, z_avg, const)
            self.species_couplings[i] = sp.coupling
            self.coupling_constant += sp.concentration * sp.coupling

    def pretty_print(self, electron_properties: bool = True):
        """
        Print simulation parameters in a user-friendly way.

        Parameters
        ----------
        electron_properties: bool
            Flag for printing electron properties. Default = True

        """
        print("\nSIMULATION BOX:")
        print('Units: ', self.units)
        print('Wigner-Seitz radius = {:.6e} '.format(self.a_ws), end='')
        print("[cm]" if self.units == "cgs" else "[m]")
        print('No. of non-zero box dimensions = ', int(self.dimensions))
        print('Box side along x axis = {:.6e} a_ws = {:.6e} '.format(
            self.box_lengths[0] / self.a_ws, self.box_lengths[0]),
            end='')
        print("[cm]" if self.units == "cgs" else "[m]")

        print('Box side along y axis = {:.6e} a_ws = {:.6e} '.format(
            self.box_lengths[1] / self.a_ws, self.box_lengths[1]),
            end='')
        print("[cm]" if self.units == "cgs" else "[m]")

        print('Box side along z axis = {:.6e} a_ws = {:.6e} '.format(
            self.box_lengths[2] / self.a_ws, self.box_lengths[2]),
            end='')
        print("[cm]" if self.units == "cgs" else "[m]")
        print("Box Volume = {:.6e} ".format(self.box_volume), end='')
        print("[cm^3]" if self.units == "cgs" else "[m^3]")

        print('Boundary conditions: {}'.format(self.boundary_conditions))

        if electron_properties:
            print('\nELECTRON PROPERTIES:')
            print('Number density: n_e = {:.6e} '.format(self.ne), end='')
            print("[N/cc]" if self.units == "cgs" else "[N/m^3]")

            print('Wigner-Seitz radius: a_e = {:.6e} '.format(self.ae_ws), end='')
            print("[cm]" if self.units == "cgs" else "[m]")

            print('Temperature: T_e = {:.6e} [K] = {:.6e} [eV]'.format(
                self.electron_temperature, self.electron_temperature / self.eV2K))

            print('de Broglie wavelength: lambda_deB = {:.6e} '.format(self.lambda_deB), end='')
            print("[cm]" if self.units == "cgs" else "[m]")

            print('Thomas-Fermi length: lambda_TF = {:.6e} '.format(self.lambda_TF), end='')
            print("[cm]" if self.units == "cgs" else "[m]")

            print('Fermi wave number: k_F = {:.6e} '.format(self.kF), end='')
            print("[1/cm]" if self.units == "cgs" else "[1/m]")

            print('Fermi Energy: E_F = {:.6e} [eV]'.format(self.fermi_energy / self.kB / self.eV2K))

            print("Relativistic parameter: x_F = {:.6e}".format(self.relativistic_parameter), end='')
            kf_xf = self.me * self.c0 ** 2 * (np.sqrt(1.0 + self.relativistic_parameter ** 2) - 1.0)
            print(" --> E_F = {:.6e} [eV]".format(kf_xf / self.kB / self.eV2K))

            print('Degeneracy parameter: Theta = {:.6e} '.format(self.electron_degeneracy_parameter))
            print('Coupling: r_s = {:.6f},  Gamma_e = {:.6f}'.format(self.rs, self.electron_coupling) )
            print('Warm Dense Matter Parameter: W = {:.4e}'.format(self.wdm_parameter))

            print("Chemical potential: mu = {:.4e} k_B T_e = {:.4e} E_F".format(
                self.eta_e, self.eta_e * self.kB * self.electron_temperature / self.fermi_energy))

            if self.magnetized:
                print("Electron cyclotron frequency: w_c = {:.6e}".format(self.electron_cyclotron_frequency))
                print("Lowest Landau energy level: h w_c/2 = {:.6e}".format(self.electron_magnetic_energy/2))
                print("Electron magnetic energy gap: h w_c = {:.6e} = {:.4e} E_F = {:.4e} k_B T_e".format(
                    self.electron_magnetic_energy,
                    self.electron_magnetic_energy/self.fermi_energy,
                    self.electron_magnetic_energy / (self.kB * self.electron_temperature),
                ))

        if self.magnetized:
            print('\nMAGNETIC FIELD:')
            print('Magnetic Field = [{:.4e}, {:.4e}, {:.4e}] '.format(*self.magnetic_field), end='')
            print("[Tesla]" if self.units == "mks" else "[Gauss]")
            print('Magnetic Field Magnitude = {:.4e} '.format(np.linalg.norm(self.magnetic_field)), end='')
            print("[Tesla]" if self.units == "mks" else "[Gauss]")
            print('Magnetic Field Unit Vector = [{:.4e}, {:.4e}, {:.4e}]'.format(
                *self.magnetic_field/np.linalg.norm(self.magnetic_field) ) )


class Particles:
    """
    Class handling particles' properties.

    Attributes
    ----------
    kB : float
        Boltzmann constant.

    pos : numpy.ndarray
        Particles' positions.

    vel : numpy.ndarray
        Particles' velocities.

    acc : numpy.ndarray
        Particles' accelerations.

    box_lengths : numpy.ndarray
        Box sides' lengths.

    masses : numpy.ndarray
        Mass of each particle. Shape = (``total_num_ptcls``).

    charges : numpy.ndarray
        Charge of each particle. Shape = (``total_num_ptcls``).

    id : numpy.ndarray,
        Species identifier. Shape = (``total_num_ptcls``).

    names : numpy.ndarray
        Species' names. Shape = (``total_num_ptcls``).

    rdf_nbins : int
        Number of bins for radial pair distribution.

    no_grs : int
        Number of independent :math:`g_{ij}(r)`.

    rdf_hist : numpy.ndarray
        Histogram array for the radial pair distribution function.

    prod_dump_dir : str
        Directory name where to store production phase's simulation's checkpoints. Default = 'dumps'.

    eq_dump_dir : str
        Directory name where to store equilibration phase's simulation's checkpoints. Default = 'dumps'.

    total_num_ptcls : int
        Total number of simulation's particles.

    num_species : int
        Number of species.

    species_num : numpy.ndarray
        Number of particles of each species. Shape = ``num_species``.

    dimensions : int
        Number of non-zero dimensions. Default = 3.

    potential_energy : float
        Instantaneous value of the potential energy.

    rnd_gen : numpy.random.Generator
        Random number generator.

    """

    def __init__(self):
        self.potential_energy = 0.0
        self.kB = None
        self.prod_dump_dir = None
        self.eq_dump_dir = None
        self.box_lengths = None
        self.total_num_ptcls = None
        self.num_species = None
        self.species_num = None
        self.dimensions = None
        self.rnd_gen = None

        self.pos = None
        self.vel = None
        self.acc = None

        self.pbc_cntr = None

        self.names = None
        self.id = None

        self.species_init_vel = None
        self.species_thermal_velocity = None

        self.masses = None
        self.charges = None
        self.cyclotron_frequencies = None

        self.no_grs = None
        self.rdf_hist = None

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = 'Particles( \n'
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ')'
        return disp

    def setup(self, params, species):
        """
        Initialize class' attributes

        Parameters
        ----------
        params: sarkas.core.Parameters
            Simulation's parameters.

        species : list
            List of :meth:`sarkas.core.Species` objects.

        """

        self.kB = params.kB
        self.prod_dump_dir = params.prod_dump_dir
        self.eq_dump_dir = params.eq_dump_dir
        self.box_lengths = np.copy(params.box_lengths)
        self.total_num_ptcls = params.total_num_ptcls
        self.num_species = params.num_species
        self.species_num = np.copy(params.species_num)
        self.dimensions = params.dimensions

        if hasattr(params, "rand_seed"):
            self.rnd_gen = np.random.Generator(np.random.PCG64(params.rand_seed))
        else:
            self.rnd_gen = np.random.Generator(np.random.PCG64(123456789))

        self.pos = np.zeros((self.total_num_ptcls, params.dimensions))
        self.vel = np.zeros((self.total_num_ptcls, params.dimensions))
        self.acc = np.zeros((self.total_num_ptcls, params.dimensions))

        self.pbc_cntr = np.zeros((self.total_num_ptcls, params.dimensions))

        self.names = np.empty(self.total_num_ptcls, dtype=params.species_names.dtype)
        self.id = np.zeros(self.total_num_ptcls, dtype=int)

        self.species_init_vel = np.zeros((params.num_species, 3))
        self.species_thermal_velocity = np.zeros((params.num_species, 3))

        self.masses = np.zeros(self.total_num_ptcls)  # mass of each particle
        self.charges = np.zeros(self.total_num_ptcls)  # charge of each particle
        self.cyclotron_frequencies = np.zeros(self.total_num_ptcls)
        # No. of independent rdf
        self.no_grs = int(self.num_species * (self.num_species + 1) / 2)
        if hasattr(params, 'rdf_nbins'):
            self.rdf_nbins = params.rdf_nbins
        else:
            # nbins = 5% of the number of particles.
            self.rdf_nbins = int(0.05 * params.total_num_ptcls)
            params.rdf_nbins = self.rdf_nbins

        self.rdf_hist = np.zeros((self.rdf_nbins, self.num_species, self.num_species))

        self.update_attributes(species)

        self.load(params)

    def load(self, params):
        """
        Initialize particles' positions and velocities.
        Positions are initialized based on the load method while velocities are chosen
        from a Maxwell-Boltzmann distribution.

        Parameters
        ----------
        params: sarkas.core.Parameters
            Simulation's parameters.

        """
        # Particles Position Initialization
        if params.load_method in ['equilibration_restart', 'eq_restart',
                                  'magnetization_restart', 'mag_restart',
                                  'production_restart', 'prod_restart']:
            # checks
            assert hasattr(params, 'restart_step'), "Restart step not defined. Please define restart_step."
            assert type(params.restart_step) is int, "Only integers are allowed."

            if params.load_method[:2] == 'eq':
                self.load_from_restart('equilibration', params.restart_step)
            elif params.load_method[:2] == 'pr':
                self.load_from_restart('production', params.restart_step)
            elif params.load_method[:2] == 'ma':
                self.load_from_restart('magnetization', params.restart_step)

        elif params.load_method == 'file':
            msg = 'Input file not defined. Please define particle_input_file.'
            assert params.ptcls_input_file, msg
            self.load_from_file(params.ptcls_input_file)

        # position distribution.
        elif params.load_method == 'lattice':
            self.lattice(params.load_perturb)

        elif params.load_method == 'random_reject':
            assert params.load_rejection_radius, "Rejection radius not defined. Please define load_rejection_radius."
            self.random_reject(params.load_rejection_radius)

        elif params.load_method == 'halton_reject':
            assert params.load_rejection_radius, "Rejection radius not defined. Please define load_rejection_radius."
            self.halton_reject(params.load_halton_bases, params.load_rejection_radius)

        elif params.load_method in ['uniform', 'random_no_reject']:
            self.pos = self.uniform_no_reject([0.0, 0.0, 0.0], params.box_lengths)

        else:
            raise AttributeError('Incorrect particle placement scheme specified.')

    def gaussian(self, mean, sigma, num_ptcls):
        """
        Initialize particles' velocities according to a normalized Maxwell-Boltzmann (Normal) distribution.
        It calls ``numpy.random.Generator.normal``

        Parameters
        ----------
        num_ptcls : int
            Number of particles to initialize.

        mean : float
            Center of the normal distribution.

        sigma : float
            Scale of the normal distribution.

        Returns
        -------
         : numpy.ndarray
            Particles property distributed according to a Normal probability density function.

        """
        return self.rnd_gen.normal(mean, sigma, (num_ptcls, 3))

    def update_attributes(self, species):
        """
        Assign particles attributes.

        Parameters
        ----------
        species : list
            List of ``sarkas.core.Species`` objects.

        """
        species_end = 0
        for ic, sp in enumerate(species):
            species_start = species_end
            species_end += sp.num

            self.names[species_start:species_end] = sp.name
            self.masses[species_start:species_end] = sp.mass

            if hasattr(sp, 'charge'):
                self.charges[species_start:species_end] = sp.charge
            else:
                self.charges[species_start:species_end] = 1.0

            if hasattr(sp, 'cyclotron_frequency'):
                self.cyclotron_frequencies[species_start:species_end] = sp.cyclotron_frequency

            self.id[species_start:species_end] = ic

            if hasattr(sp, "init_vel"):
                self.species_init_vel[ic, :] = sp.init_vel

            if sp.initial_velocity_distribution == "boltzmann":
                if isinstance(sp.temperature, (int, float)):
                    sp_temperature = np.ones(self.dimensions) * sp.temperature

                self.species_thermal_velocity[ic] = np.sqrt(self.kB * sp_temperature / sp.mass)
                self.vel[species_start:species_end, :] = self.gaussian(sp.initial_velocity,
                                                                       self.species_thermal_velocity[ic], sp.num)

    def load_from_restart(self, phase, it):
        """
        Load particles' data from a checkpoint of a previous run

        Parameters
        ----------
        it : int
            Timestep.

        phase: str
            Restart phase.

        """
        if phase == 'equilibration':
            file_name = os.path.join(self.eq_dump_dir, "checkpoint_" + str(it) + ".npz")
            data = np.load(file_name, allow_pickle=True)
            self.id = data["id"]
            self.names = data["names"]
            self.pos = data["pos"]
            self.vel = data["vel"]
            self.acc = data["acc"]

        elif phase == 'production':
            file_name = os.path.join(self.prod_dump_dir, "checkpoint_" + str(it) + ".npz")
            data = np.load(file_name, allow_pickle=True)
            self.id = data["id"]
            self.names = data["names"]
            self.pos = data["pos"]
            self.vel = data["vel"]
            self.acc = data["acc"]
            self.pbc_cntr = data["cntr"]
            self.rdf_hist = data["rdf_hist"]

        elif phase == 'magnetization':
            file_name = os.path.join(self.mag_dump_dir, "checkpoint_" + str(it) + ".npz")
            data = np.load(file_name, allow_pickle=True)
            self.id = data["id"]
            self.names = data["names"]
            self.pos = data["pos"]
            self.vel = data["vel"]
            self.acc = data["acc"]
            self.pbc_cntr = data["cntr"]
            self.rdf_hist = data["rdf_hist"]

    def load_from_file(self, f_name):
        """
        Load particles' data from a specific file.

        Parameters
        ----------
        f_name : str
            Filename
        """
        pv_data = np.loadtxt(f_name)
        if not (pv_data.shape[0] == self.total_num_ptcls):
            print("Number of particles is not same between input file and initial p & v data file.")
            print("From the input file: N = ", self.total_num_ptcls)
            print("From the initial p & v data: N = ", pv_data.shape[0])
            sys.exit()
        self.pos[:, 0] = pv_data[:, 0]
        self.pos[:, 1] = pv_data[:, 1]
        self.pos[:, 2] = pv_data[:, 2]

        self.vel[:, 0] = pv_data[:, 3]
        self.vel[:, 1] = pv_data[:, 4]
        self.vel[:, 2] = pv_data[:, 5]

    def uniform_no_reject(self, mins, maxs):
        """
        Randomly distribute particles along each direction.

        Parameters
        ----------
        mins : float
            Minimum value of the range of a uniform distribution.

        maxs : float
            Maximum value of the range of a uniform distribution.

        Returns
        -------
         : numpy.ndarray
            Particles' property, e.g. pos, vel. Shape = (``total_num_ptcls``, 3).

        """

        return self.rnd_gen.uniform(mins, maxs, (self.total_num_ptcls, 3))

    def lattice(self, perturb):
        """
        Place particles in a simple cubic lattice with a slight perturbation ranging
        from 0 to 0.5 times the lattice spacing.

        Parameters
        ----------
        perturb : float
            Value of perturbation, p, such that 0 <= p <= 1.

        Notes
        -----
        Author: Luke Stanek
        Date Created: 5/6/19
        Date Updated: 6/2/19
        Updates: Added to S_init_schemes.py for Sarkas import
        """

        # Check if perturbation is below maximum allowed. If not, default to maximum perturbation.
        if perturb > 1:
            print('Warning: Random perturbation must not exceed 1. Setting perturb = 1.')
            perturb = 1  # Maximum perturbation

        print('Initializing particles with maximum random perturbation of {} times the lattice spacing.'.format(
            perturb * 0.5))

        # Determining number of particles per side of simple cubic lattice
        part_per_side = self.total_num_ptcls ** (1. / 3.)  # Number of particles per side of cubic lattice

        # Check if total number of particles is a perfect cube, if not, place more than the requested amount
        if round(part_per_side) ** 3 != self.total_num_ptcls:
            part_per_side = np.ceil(self.total_num_ptcls ** (1. / 3.))
            print('\nWARNING: Total number of particles requested is not a perfect cube.')
            print('Initializing with {} particles.'.format(int(part_per_side ** 3)))

        dx_lattice = self.box_lengths[0] / (self.total_num_ptcls ** (1. / 3.))  # Lattice spacing
        dz_lattice = self.box_lengths[1] / (self.total_num_ptcls ** (1. / 3.))  # Lattice spacing
        dy_lattice = self.box_lengths[2] / (self.total_num_ptcls ** (1. / 3.))  # Lattice spacing

        # Create x, y, and z position arrays
        x = np.arange(0, self.box_lengths[0], dx_lattice) + 0.5 * dx_lattice
        y = np.arange(0, self.box_lengths[1], dy_lattice) + 0.5 * dy_lattice
        z = np.arange(0, self.box_lengths[2], dz_lattice) + 0.5 * dz_lattice

        # Create a lattice with appropriate x, y, and z values based on arange
        X, Y, Z = np.meshgrid(x, y, z)

        # Perturb lattice
        X += self.rnd_gen.uniform(-0.5, 0.5, np.shape(X)) * perturb * dx_lattice
        Y += self.rnd_gen.uniform(-0.5, 0.5, np.shape(Y)) * perturb * dy_lattice
        Z += self.rnd_gen.uniform(-0.5, 0.5, np.shape(Z)) * perturb * dz_lattice

        # Flatten the meshgrid values for plotting and computation
        self.pos[:, 0] = X.ravel()
        self.pos[:, 1] = Y.ravel()
        self.pos[:, 2] = Z.ravel()

    def random_reject(self, r_reject):
        """
        Place particles by sampling a uniform distribution from 0 to L (the box length)
        and uses a rejection radius to avoid placing particles to close to each other.

        Parameters
        ----------
        r_reject : float
            Value of rejection radius.

        Notes
        -----
        Author: Luke Stanek
        Date Created: 5/6/19
        Date Updated: N/A
        Updates: N/A

        """

        # Initialize Arrays
        x = np.array([])
        y = np.array([])
        z = np.array([])

        # Set first x, y, and z positions
        x_new = self.rnd_gen.uniform(0, self.box_lengths[0])
        y_new = self.rnd_gen.uniform(0, self.box_lengths[1])
        z_new = self.rnd_gen.uniform(0, self.box_lengths[2])

        # Append to arrays
        x = np.append(x, x_new)
        y = np.append(y, y_new)
        z = np.append(z, z_new)

        # Particle counter
        i = 0

        cntr_reject = 0
        cntr_total = 0
        # Loop to place particles
        while i < self.total_num_ptcls - 1:

            # Set x, y, and z positions
            x_new = self.rnd_gen.uniform(0.0, self.box_lengths[0])
            y_new = self.rnd_gen.uniform(0.0, self.box_lengths[1])
            z_new = self.rnd_gen.uniform(0.0, self.box_lengths[2])

            # Check if particle was place too close relative to all other current particles
            for j in range(len(x)):

                # Flag for if particle is outside of cutoff radius (True -> not inside rejection radius)
                flag = 1

                # Compute distance b/t particles for initial placement
                x_diff = x_new - x[j]
                y_diff = y_new - y[j]
                z_diff = z_new - z[j]

                # periodic condition applied for minimum image
                if x_diff < - self.box_lengths[0] / 2:
                    x_diff += self.box_lengths[0]
                if x_diff > self.box_lengths[0] / 2:
                    x_diff -= self.box_lengths[0]

                if y_diff < - self.box_lengths[1] / 2:
                    y_diff += self.box_lengths[1]
                if y_diff > self.box_lengths[1] / 2:
                    y_diff -= self.box_lengths[1]

                if z_diff < -self.box_lengths[2] / 2:
                    z_diff += self.box_lengths[2]
                if z_diff > self.box_lengths[2] / 2:
                    z_diff -= self.box_lengths[2]

                # Compute distance
                r = np.sqrt(x_diff ** 2 + y_diff ** 2 + z_diff ** 2)

                # Check if new particle is below rejection radius. If not, break out and try again
                if r <= r_reject:
                    flag = 0  # new position not added (False -> no longer outside reject r)
                    cntr_reject += 1
                    cntr_total += 1
                    break

            # If flag true add new position
            if flag == 1:
                x = np.append(x, x_new)
                y = np.append(y, y_new)
                z = np.append(z, z_new)

                # Increment particle number
                i += 1
                cntr_total += 1

        self.pos[:, 0] = x
        self.pos[:, 1] = y
        self.pos[:, 2] = z

    def halton_reject(self, bases, r_reject):
        """
        Place particles according to a Halton sequence from 0 to L (the box length)
        and uses a rejection radius to avoid placing particles to close to each other.

        Parameters
        ----------
        bases : numpy.ndarray
            Array of 3 ints each of which is a base for the Halton sequence.
            Defualt: bases = np.array([2,3,5])

        r_reject : float
            Value of rejection radius.

        Notes
        -----
        Author: Luke Stanek
        Date Created: 5/6/19
        Date Updated: N/A
        Updates: N/A

        """

        # Get bases
        b1, b2, b3 = bases

        # Allocate space and store first value from Halton
        x = np.array([0])
        y = np.array([0])
        z = np.array([0])

        # Initialize particle counter and Halton counter
        i = 1
        k = 1

        # Loop over all particles
        while i < self.total_num_ptcls:

            # Increment particle counter
            n = k
            m = k
            p = k

            # Determine x coordinate
            f1 = 1
            r1 = 0
            while n > 0:
                f1 /= b1
                r1 += f1 * (n % int(b1))
                n = np.floor(n / b1)
            x_new = self.box_lengths[0] * r1  # new x value

            # Determine y coordinate
            f2 = 1
            r2 = 0
            while m > 0:
                f2 /= b2
                r2 += f2 * (m % int(b2))
                m = np.floor(m / b2)
            y_new = self.box_lengths[1] * r2  # new y value

            # Determine z coordinate
            f3 = 1
            r3 = 0
            while p > 0:
                f3 /= b3
                r3 += f3 * (p % int(b3))
                p = np.floor(p / b3)
            z_new = self.box_lengths[2] * r3  # new z value

            # Check if particle was place too close relative to all other current particles
            for j in range(len(x)):

                # Flag for if particle is outside of cutoff radius (1 -> not inside rejection radius)
                flag = 1

                # Compute distance b/t particles for initial placement
                x_diff = x_new - x[j]
                y_diff = y_new - y[j]
                z_diff = z_new - z[j]

                # Periodic condition applied for minimum image
                if x_diff < - self.box_lengths[0] / 2:
                    x_diff = x_diff + self.box_lengths[0]
                if x_diff > self.box_lengths[0] / 2:
                    x_diff = x_diff - self.box_lengths[0]

                if y_diff < -self.box_lengths[1] / 2:
                    y_diff = y_diff + self.box_lengths[1]
                if y_diff > self.box_lengths[1] / 2:
                    y_diff = y_diff - self.box_lengths[1]

                if z_diff < -self.box_lengths[2] / 2:
                    z_diff = z_diff + self.box_lengths[2]
                if z_diff > self.box_lengths[2] / 2:
                    z_diff = z_diff - self.box_lengths[2]

                # Compute distance
                r = np.sqrt(x_diff ** 2 + y_diff ** 2 + z_diff ** 2)

                # Check if new particle is below rejection radius. If not, break out and try again
                if r <= r_reject:
                    k += 1  # Increment Halton counter
                    flag = 0  # New position not added (0 -> no longer outside reject r)
                    break

            # If flag true add new positiion
            if flag == 1:
                # Add new positions to arrays
                x = np.append(x, x_new)
                y = np.append(y, y_new)
                z = np.append(z, z_new)

                k += 1  # Increment Halton counter
                i += 1  # Increment particle number

        self.pos[:, 0] = x
        self.pos[:, 1] = y
        self.pos[:, 2] = z

    def kinetic_temperature(self):
        """
        Calculate the kinetic energy and temperature of each species.

        Returns
        -------
        K : numpy.ndarray
            Kinetic energy of each species. Shape=(``num_species``).

        T : numpy.ndarray
            Temperature of each species. Shape=(``num_species``).

        """
        K = np.zeros(self.num_species)
        T = np.zeros(self.num_species)
        const = 2.0 / (self.kB * self.species_num * self.dimensions)
        kinetic = 0.5 * self.masses * (self.vel * self.vel).transpose()

        species_start = 0
        species_end = 0
        for i, num in enumerate(self.species_num):
            species_end += num
            K[i] = np.sum(kinetic[:, species_start:species_end])
            T[i] = const[i] * K[i]
            species_start = species_end

        return K, T

    def remove_drift(self):
        """
        Enforce conservation of total linear momentum. Updates particles velocities
        """
        species_start = 0
        species_end = 0
        momentum = self.masses * self.vel.transpose()
        for ic, nums in enumerate(self.species_num):
            species_end += nums
            P = np.sum(momentum[:, species_start:species_end], axis=1)
            self.vel[species_start:species_end, :] -= P / (nums * self.masses[species_end - 1])
            species_start = species_end


class Species:
    """
    Class used to store all the information of a single species.

    Attributes
    ----------
    name : str
        Species' name.

    number_density : float
        Species number density in appropriate units.

    num : int
        Number of particles of Species.

    mass : float
        Species' mass.

    charge : float
        Species' charge.

    Z : float
        Species charge number or ionization degree.

    ai : float
        Species Wigner - Seitz radius calculate from electron density.
        Used in the calculation of the effective coupling constant.

    ai_dens : float
        Species Wigner - Seitz radius calculate from the species density.

    coupling : float
        Species coupling constant

    plasma_frequency : float
        Species' plasma frequency.

    debye_length : float
        Species' Debye Length.

    cyclotron_frequency : float
        Species' cyclotron frequency.

    initial_velocity : numpy.ndarray
        Initial velocity in x,y,z directions.

    temperature : float
        Initial temperature of the species.

    temperature_eV : float
        Initial temperature of the species in eV.

    initial_velocity_distribution : str
        Type of distribution. Default = 'boltzmann'.

    initial_spatial_distribution : str
        Type of distribution. Default = 'uniform'.

    atomic_weight : float
        (Optional) Species mass in atomic units.

    concentration : float
        Species' concentration.

    mass_density : float
        (Optional) Species' mass density.


    """

    def __init__(self, input_dict: dict = None):
        """
        Parameters
        ----------
        input_dict: dict, optional
            Dictionary with species information.

        """
        self.ai = None
        self.ai_dens = None
        self.atomic_weight = None
        self.charge = None
        self.Z = None
        self.coupling = None
        self.concentration = None
        self.debye_length = None
        self.initial_spatial_distribution = 'random_no_reject'
        self.initial_velocity_distribution = 'boltzmann'
        self.initial_velocity = np.zeros(3)
        self.mass = None
        self.mass_density = None
        self.name = None
        self.num = None
        self.number_density = None
        self.cyclotron_frequency = None
        self.plasma_frequency = None
        self.temperature = None
        self.temperature_eV = None

        if input_dict:
            self.from_dict(input_dict)

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = 'Species( \n'
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ')'
        return disp

    def from_dict(self, input_dict: dict):
        """
        Update attributes from input dictionary using the ``__dict__.update()`` method.

        Parameters
        ----------
        input_dict: dict
            Dictionary to be copied.

        """
        self.__dict__.update(input_dict)
        if not isinstance(self.initial_velocity, np.ndarray):
            self.initial_velocity = np.array(self.initial_velocity)

    def calc_plasma_frequency(self, constant: float):
        """
        Calculate the plasma frequency.

        Parameters
        ----------
        constant : float
            Charged systems: Electrostatic constant  :math: `4\pi \epsilon_0` [mks]
            Neutral systems: :math: `1/n\sigma^2`

        """
        self.plasma_frequency = np.sqrt(4.0 * np.pi * self.charge ** 2 * self.number_density / (self.mass * constant))

    def calc_debye_length(self, kB: float, constant: float):
        """
        Calculate the Debye Length.

        Parameters
        ----------
        kB : float
            Boltzmann constant.

        constant : float
            Charged systems: Electrostatic constant  :math: `4 \pi \epsilon_0` [mks]
            Neutral systems: :math: `1/n\sigma^2`

        """
        self.debye_length = np.sqrt((self.temperature * kB * constant)
                                    / (4.0 * np.pi * self.charge ** 2 * self.number_density))

    def calc_cyclotron_frequency(self, magnetic_field_strength: float):
        """
        Calculate the cyclotron frequency. See `Wikipedia link <https://en.wikipedia.org/wiki/Lorentz_force>`_.

        Parameters
        ----------

        magnetic_field_strength : float
            Magnetic field strength.

        """
        self.cyclotron_frequency = self.charge * magnetic_field_strength / self.mass

    def calc_coupling(self, a_ws: float, z_avg: float, const: float):
        """
        Calculate the coupling constant between particles.

        Parameters
        ----------
        a_ws : float
            Total Wigner-Seitz radius.

        z_avg : float
            Average charge of the system.

        const : float
            Electrostatic * Thermal constants.

        """
        self.ai_dens = (3.0 / (4.0 * np.pi * self.number_density)) ** (1. / 3.)
        self.ai = (self.charge / z_avg) ** (1. / 3.) * a_ws if z_avg > 0 else self.ai_dens
        self.coupling = self.charge ** 2 / (self.ai * const * self.temperature)

    def pretty_print(self, potential_type: str = None, units: str = 'mks'):
        """Print Species information in a user-friendly way.

        Parameters
        ----------
        potential_type: str
            Interaction potential. If 'LJ' it will print out the epsilon and sigma attributes.

        units: str
            Unit system used in the simulation. Default = 'mks'.

        """

        print("\tName: {}".format(self.name))
        print("\tNo. of particles = {} ".format(self.num))
        print("\tNumber density = {:.6e} ".format(self.number_density), end='')
        print("[N/cc]" if units == "cgs" else "[N/m^3]")
        print("\tAtomic weight = {:.4f} [a.u.]".format(self.atomic_weight))
        print("\tMass = {:.6e} ".format(self.mass), end='')
        print("[g]" if units == "cgs" else "[kg]")
        print("\tMass density = {:.6e} ".format(self.mass_density), end='')
        print("[g/cc]" if units == "cgs" else "[kg/m^3]")
        print("\tCharge number/ionization degree = {:.4f} ".format(self.Z))
        print("\tCharge = {:.6e} ".format(self.charge), end='')
        print("[esu]" if units == "cgs" else "[C]")
        print('\tTemperature = {:.6e} [K] = {:.6e} [eV]'.format(self.temperature, self.temperature_eV))
        if potential_type == 'LJ':
            print("\tEpsilon = {:.6e} ".format(self.epsilon), end='')
            print("[erg]" if units == "cgs" else "[J]")
            print("\tSigma = {:.6e} ".format(self.sigma), end='')
            print("[cm]" if units == "cgs" else "[m]")

        print('\tDebye Length = {:.6e} [Hz]'.format(self.debye_length))
        print('\tPlasma Frequency = {:.6e} [Hz]'.format(self.plasma_frequency))
        if self.cyclotron_frequency:
            print('\tCyclotron Frequency = {:.6e} [Hz]'.format(self.cyclotron_frequency) )
            print('\tbeta_c = {:.4f}'.format(self.cyclotron_frequency/self.plasma_frequency))

    # Methods
    # -------
    # from_dict(input_dict: dict)
    #     Update attributes from input dictionary.
    #
    # calc_plasma_frequency(constant: float)
    #     Calculate the species' plasma frequency.
    #
    # calc_debye_length(kb: float, constant: float)
    #     Calculate the species' Debye length.
    #
    # calc_cyclotron_frequency(magnetic_field_strength: float)
    #     Calculate the cyclotron frequency.
    #
    # calc_coupling(a_ws: float, z_avg: float, const:float)
    #     Calculate the species' Wigner-Seitz radius and use it to calculate the
    #     coupling constant of the species.



"""
Module containing various thermostat. Berendsen only for now.
"""
import numpy as np
from numba import njit


class Thermostat:
    """
    Thermostat object.

    Attributes
    ----------
    kB : float
        Boltzmann constant in correct units.

    no_species : int
        Total number of species.

    species_np : numpy.ndarray
        Number of particles of each species.

    species_masses : numpy.ndarray
        Mass of each species.

    relaxation_rate: float
        Berendsen parameter tau.

    relaxation_timestep: int
        Timestep at which thermostat is turned on.

    type: str
        Thermostat type

    berendsen_tau: float
        Berendsen parameter.

    """

    def __init__(self):
        self.temperatures = None
        self.temperatures_eV = None
        self.type = None
        self.relaxation_rate = None
        self.relaxation_timestep = None
        self.kB = None
        self.species_num = None
        self.species_masses = None
        self.berendsen_tau = None
        self.eV_temp_flag = False
        self.K_temp_flag = False

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = 'Thermostat( \n'
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ')'
        return disp

    def from_dict(self, input_dict: dict):
        """
        Update attributes from input dictionary.

        Parameters
        ----------
        input_dict: dict
            Dictionary to be copied.

        """
        self.__dict__.update(input_dict)

        # Make sure list are turned into numpy arrays
        if self.temperatures_eV:
            if not isinstance(self.temperatures_eV, np.ndarray):
                self.temperatures_eV = np.array([self.temperatures_eV])
            self.eV_temp_flag = True

        if self.temperatures:
            if not isinstance(self.temperatures, np.ndarray):
                self.temperatures = np.array([self.temperatures])
            self.K_temp_flag = True

    def pretty_print(self):
        """Print Thermostat information in a user-friendly way."""
        print('Type: {}'.format(self.type))
        print('First thermostating timestep, i.e. relaxation_timestep = {}'.format(self.relaxation_timestep))
        print("Berendsen parameter tau: {:.3f} [timesteps]".format(self.berendsen_tau))
        print("Berendsen relaxation rate: {:.3f} [1/timesteps] ".format(self.relaxation_rate))
        if not self.eV_temp_flag and not self.K_temp_flag:
            # If you forgot to give thermostating temperatures
            print("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! WARNING !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("Equilibration temperatures not defined. I will use the species's temperatures")
            print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n")
        print("Thermostating temperatures: ")
        for i, (t, t_ev) in enumerate(zip(self.temperatures, self.temperatures_eV)):
            print("Species ID {}: T_eq = {:.6e} [K] = {:.6e} [eV]".format(i, t, t_ev))

    def setup(self, params):
        """
        Assign attributes from simulation's parameters.

        Parameters
        ----------
        params: sarkas.core.Parameters
            Simulation's parameters

        """

        # Check whether you input temperatures in eV or K

        if self.eV_temp_flag:
            self.temperatures = params.eV2K * np.copy(self.temperatures_eV)
        elif self.K_temp_flag:
            self.temperatures_eV = np.copy(self.temperatures) / params.eV2K
        else:
            self.temperatures = np.copy(params.species_temperatures)
            self.temperatures_eV = np.copy(self.temperatures) / params.eV2K

        if self.berendsen_tau:
            self.relaxation_rate = 1.0 / self.berendsen_tau
        else:
            self.berendsen_tau = 1.0 / self.relaxation_rate

        if not self.temperatures.all():
            self.temperatures = np.copy(params.species_temperatures)

        self.kB = params.kB
        self.species_num = np.copy(params.species_num)
        self.species_masses = np.copy(params.species_masses)

        assert self.type.lower() == "berendsen", "Only Berendsen thermostat is supported."

    def update(self, ptcls, it):
        """
        Update particles' velocities according to the chosen thermostat

        Parameters
        ----------
        ptcls : sarkas.core.Particles
            Particles' data.

        it : int
            Current timestep.

        """
        K, T = ptcls.kinetic_temperature()
        berendsen(ptcls.vel, self.temperatures, T, self.species_num, self.relaxation_timestep,
                  self.relaxation_rate, it)


@njit
def calc_kin_temp(vel, nums, masses, kB):
    """
    Calculates the kinetic energy and temperature.

    Parameters
    ----------
    kB: float
        Boltzmann constant in chosen units.

    masses: numpy.ndarray
        Mass of each species.

    nums: numpy.ndarray
        Number of particles of each species.

    vel: numpy.ndarray
        Particles' velocities.

    Returns
    -------
    K : numpy.ndarray
        Kinetic energy of each species.

    T : numpy.ndarray
        Temperature of each species.
    """

    num_species = len(nums)

    K = np.zeros(num_species)
    T = np.zeros(num_species)
    const = 2.0 / (kB * nums * vel.shape[1])
    kinetic_energies = 0.5 * masses * (vel ** 2).transpose()

    species_start = 0
    species_end = 0
    for i, num in enumerate(nums):
        species_end += num
        K[i] = np.sum(kinetic_energies[:, species_start:species_end])
        T[i] = const[i] * K[i]
        species_start += num

    return K, T


@njit
def berendsen(vel, T_desired, T, species_np, therm_timestep, tau, it):
    """
    Update particle velocity based on Berendsen thermostat [Berendsen1984]_.

    Parameters
    ----------
    T : numpy.ndarray
        Instantaneous temperature of each species.

    vel : numpy.ndarray
        Particles' velocities to rescale.

    T_desired : numpy.ndarray
        Target temperature of each species.

    tau : float
        Scale factor.

    therm_timestep : int
        Timestep at which to turn on the thermostat.

    species_np : numpy.ndarray
        Number of each species.

    it : int
        Current timestep.

    References
    ----------
    .. [Berendsen1984] `H.J.C. Berendsen et al., J Chem Phys 81 3684 (1984) <https://doi.org/10.1063/1.448118>`_

    """

    # if it < therm_timestep:
    #     fact = np.sqrt(T_desired / T)
    # else:
    #     fact = np.sqrt(1.0 + (T_desired / T - 1.0) * tau)  # eq.(11)

    # branchless programming
    fact = 1.0 * (it < therm_timestep) + np.sqrt(1.0 + (T_desired / T - 1.0) * tau) * (it >= therm_timestep)
    species_start = 0
    species_end = 0

    for i, num in enumerate(species_np):
        species_end += num
        vel[species_start:species_end, :] *= fact[i]
        species_start += num



import glob
import os
import sys
from configparser import ConfigParser
import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install


# The following are needed to copy the MSU plot styles in Matplotlib folder
# From https://stackoverflow.com/questions/20288711/post-install-script-with-python-setuptools
class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        develop.run(self)
        import matplotlib as mpl
        import shutil
        # ~ # ref  ->  matplotlib/style/core
        BASE_LIBRARY_PATH = os.path.join(mpl.get_data_path(), 'stylelib')
        STYLE_PATH = os.path.join(os.getcwd(), os.path.join('sarkas', 'mplstyles'))
        STYLE_EXTENSION = 'mplstyle'
        style_files = glob.glob(os.path.join(STYLE_PATH, "*.%s" % (STYLE_EXTENSION)))

        # Copy the plotting style in the matplotlib directory
        for _path_file in style_files:
            _, fname = os.path.split(_path_file)
            dest = os.path.join(BASE_LIBRARY_PATH, fname)
            shutil.copy(_path_file, dest)
            print("%s style installed" % (fname))


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        install.run(self)
        import matplotlib as mpl
        import shutil
        # ~ # ref  ->  matplotlib/style/core
        BASE_LIBRARY_PATH = os.path.join(mpl.get_data_path(), 'stylelib')
        STYLE_PATH = os.path.join(os.getcwd(), os.path.join('sarkas', 'mplstyles'))
        STYLE_EXTENSION = 'mplstyle'
        style_files = glob.glob(os.path.join(STYLE_PATH, "*.%s" % (STYLE_EXTENSION)))

        # Copy the plotting style in the matplotlib directory
        for _path_file in style_files:
            _, fname = os.path.split(_path_file)
            dest = os.path.join(BASE_LIBRARY_PATH, fname)
            shutil.copy(_path_file, dest)
            print("%s style installed" % (fname))


# Package Requirements
BASE_DEPENDENCIES = [
    'numpy',
    'scipy',
    'pandas',
    'numba>=0.50',
    'pyfftw',
    'fdint',
    'pyyaml',
    'tqdm',
    'pyfiglet==0.8.post1',
    'pickle5',
    'jupyter',
    'jupyterlab',
    'notebook',
    'matplotlib',
    'seaborn',
]

# Get some values from the setup.cfg
conf = ConfigParser()
conf.read(['setup.cfg'])
metadata = dict(conf.items('metadata'))

PACKAGENAME = metadata.get('package_name')
DESCRIPTION = metadata.get('description')
DESCRIPTION_FILE = metadata.get('description-file')
PACKAGEDIR = metadata.get('package_dir')
VERSION = metadata.get('version')
AUTHOR = metadata.get('author')
AUTHOR_EMAIL = metadata.get('author_email')
LICENSE = metadata.get('license')
URL = metadata.get('url')
__minimum_python_version__ = metadata.get("minimum_python_version")

# Enforce Python version check - this is the same check as in __init__.py but
# this one has to happen before importing ah_bootstrap.
if sys.version_info < tuple((int(val) for val in __minimum_python_version__.split('.'))):
    sys.stderr.write("ERROR: packagename requires Python {} or later\n".format(__minimum_python_version__))
    sys.exit(1)

# Read the README file into a string
with open(DESCRIPTION_FILE, "r") as fh:
    long_description = fh.read()

# Treat everything in scripts as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))]

setuptools.setup(
    name=PACKAGENAME,  # Replace with your own username
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    scripts=scripts,
    packages=setuptools.find_packages(),
    install_requires=BASE_DEPENDENCIES,
    # dependency_links = ['https://pypi.org/'],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    # Call the classes above and run the post installation scripts
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    },
)
